--[[=============================================================================
    Callback functions that are called by timers used for interactive functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "lib.c4_log"
require "lib.c4_timer"

StartupTimer = nil
CheckInitTimer = nil
ZoneLabelTimer = nil
TriggerPgmTimer = nil
ArmingDelayTimer = nil
TimeAndDateTimer = nil
BypassSensorTimer = nil
ManageUserCodeTimer = nil
ClearInterfaceTimer = nil

-- Time/Date flags
SetTime = false
SetDate = false
TimeShortcut = false
DateShortcut = false
ExitDateTime = false

-- Zone Label flags
ZoneIDFound = false
LabelShortcut = false
CommandShortcut = false
ZoneExitCount = 1
ZoneLabelCount = 1

-- Add User Code Flags
UserCodeEnter = false
ExitAddUserCode = false
AddUserMenuShortcut = false

function TriggerPgmCB(command)
	SendCommand(command)
	KillTimer(TriggerPgmTimer)
end

function BypassSensorCB(parms)
	local partitionId = parms["PartitionID"]
	local sensorNumber = parms["SensorID"]

	local lastIndex = PadValue(ConvertToHex(4+string.len(sensorNumber)))
	local request = lastIndex .. "40" .. PadKeyPresses(partitionId .. "0" .. sensorNumber)
	local checkSum = GetCheckSum(request)

	LogDebug("LastIndex[%s]    request[%s]    checkSum[%s]", lastIndex, request, checkSum)
	SendCommand("\n" .. request .. checkSum)
	KillTimer(BypassSensorTimer)
end

function ClearInterfaceID()
	LAST_INTERFACEID = ""
	KillTimer(ClearInterfaceTimer)
end

function ArmingFailedCB(parms)
	local partitionID = parms["PartitionID"]
	local interfaceID = parms["InterfaceID"]

	ArmPartitionFailed(partitionID, "Keypad", interfaceID)
	KillTimer(ArmingDelayTimer)
end

function StartupCB()
	FullEquipmentList()
	KillTimer(StartupTimer)
end

function CheckInitCB()
	KillTimer(CheckInitTimer)
	if (SYSTEM_INITIALIZED == false or PanelFeatures[1]["Quick Arm"] == -1) then
		EstablishCommunication()
	end
end

function SetDateAndTimeCB(DateTime)
	local dateCmd
	local timeCmd
	local request
	local checkSum
	local lastIndex

	if (ProgrammingMode == false) then
		return
	end

	-- [PN][AG][*][020(Shortcut)]
	if (TimeShortcut ~= true) then
		
		local timeCmd = "A020"
		lastIndex = PadValue(ConvertToHex(4 + string.len(timeCmd)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. timeCmd)
		checkSum = GetCheckSum(request)

		LogDebug("SetDateAndTimeCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		TimeShortcut = true

		return
	end

	-- [PN][AG][hh][mm][#]
	if (SetTime ~= true) then

		timeCmd = PadValue(DateTime.hour) .. PadValue(DateTime.minute) .. "B"
		lastIndex = PadValue(ConvertToHex(4 + string.len(timeCmd)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. timeCmd)
		checkSum = GetCheckSum(request)

		LogDebug("SetDateAndTimeCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		SetTime = true

		return
	end

	-- Continue with the time and date stuff, [*021]
	if (DateShortcut ~= true) then

		dateCmd = "A021"
		lastIndex = PadValue(ConvertToHex(4 + string.len(dateCmd)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. dateCmd)
		checkSum = GetCheckSum(request)

		LogDebug("SetDateAndTimeCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		DateShortcut = true

		return
	end

	-- Continue with the time and date stuff, [<mmddyy>,#]
	if (SetDate ~= true) then

		dateCmd = PadValue(DateTime.month) .. PadValue(DateTime.day) .. DateTime.year .. "B"
		lastIndex = PadValue(ConvertToHex(4 + string.len(dateCmd)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. dateCmd)
		checkSum = GetCheckSum(request)

		LogDebug("SetDateAndTimeCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		SetDate = true

		return
	end

	--*00#
	if (ExitDateTime ~= true) then

		local exitCmd = "A00B"
		lastIndex = PadValue(ConvertToHex(4 + string.len(exitCmd)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. exitCmd)
		checkSum = GetCheckSum(request)

		LogDebug("SetDateAndTimeCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		ExitDateTime = true
	end

	--#
	lastIndex = PadValue(ConvertToHex(5))
	request = lastIndex .. "40" .. PadKeyPresses("10B")
	checkSum = GetCheckSum(request)

	LogDebug("SetDateAndTimeCB Request: " .. request)
	SendCommand("\n" .. request .. checkSum)

	SetTime = false
	SetDate = false
	TimeShortcut = false
	DateShortcut = false
	ExitDateTime = false

	KillTimer(TimeAndDateTimer)
end

function ZoneLabelCB(Labels)
	local request
	local checkSum
	local lastIndex
	local zoneLabel

	if (CommandShortcut ~= true) then

		lastIndex = PadValue(ConvertToHex(6))
		request = lastIndex .. "40" .. PadKeyPresses("1000")
		checkSum = GetCheckSum(request)

		LogDebug("ZoneLabelCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		CommandShortcut = true

		return
	end

	if (ProgrammingMode == false) then
		return
	end

	if (LabelShortcut ~= true) then

		local shortcutCmd = "081B"
		lastIndex = PadValue(ConvertToHex(4 + string.len(shortcutCmd)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. shortcutCmd)
		checkSum = GetCheckSum(request)

		LogDebug("ZoneLabelCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		LabelShortcut = true

		return
	end

	if (ZoneIDFound ~= true) then -- Like this

		local zoneIDShortcut = "0" .. tostring(Labels[0]) .. "BB"
		lastIndex = PadValue(ConvertToHex(4 + string.len(zoneIDShortcut)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. zoneIDShortcut)
		checkSum = GetCheckSum(request)

		LogDebug("ZoneLabelCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		ZoneIDFound = true

		return
	end

	if (ZoneLabelCount <= 16) then

		if (ZoneLabelCount < tonumber(Labels["COUNT"])) then
			zoneLabel = tostring(Labels[ZoneLabelCount]) .. "B"
		else
			zoneLabel = FindLabel(" ") .. "B"
		end

		lastIndex = PadValue(ConvertToHex(4 + string.len(zoneLabel)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. zoneLabel)

		LogDebug("ZoneLabelCB Request:  ZONELABELCOUNT[" .. tostring(ZoneLabelCount) .. "]    COUNT[" .. Labels["COUNT"] .. "]")
		LogDebug("ZoneLabelCB Request: " .. request)
		checkSum = GetCheckSum(request)

		SendCommand("\n" .. request .. checkSum)
		ZoneLabelCount = tonumber(ZoneLabelCount) + 1
		
		return
	end

	if (ZoneExitCount <= 4) then

		lastIndex = PadValue(ConvertToHex(5))
		request = lastIndex .. "40" .. PadKeyPresses("10A")
		checkSum = GetCheckSum(request)

		LogDebug("ZoneLabelCB Request:  ZONEEXITCOUNT[" .. tostring(ZoneExitCount) .. "]")
		LogDebug("ZoneLabelCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		ZoneExitCount = tonumber(ZoneExitCount) + 1
		
		return
	end

	if (ZoneExitCount == 5) then

		lastIndex = PadValue(ConvertToHex(5))
		request = lastIndex .. "40" .. PadKeyPresses("10") .. "2C"
		checkSum = GetCheckSum(request)

		LogDebug("ZoneLabelCB Request:  ZONEEXITCOUNT[" .. tostring(ZoneExitCount) .. "]")
		LogDebug("ZoneLabelCB Request: " .. request)
		SendCommand("\n" .. request .. checkSum)
		ZoneExitCount = tonumber(ZoneExitCount) + 1
		
		return
	end

	lastIndex = PadValue(ConvertToHex(5))
	request = lastIndex .. "40" .. PadKeyPresses("10B")
	LogDebug("ZoneLabelCB Request: " .. request)
	checkSum = GetCheckSum(request)
	SendCommand("\n" .. request .. checkSum)

	ZoneIDFound = false
	LabelShortcut = false
	CommandShortcut = false
	ZoneExitCount = 1
	ZoneLabelCount = 1

	KillTimer(ZoneLabelTimer)
end

function ManageUserCodeCB(parms)
	local request
	local checkSum
	local lastIndex
	local zoneLabel

	if (ProgrammingMode == false) then
		return
	end

	-- [PN][AG][*][030NNN0]
	if (AddUserMenuShortcut ~= true) then
		local userNumber = PadZeros(parms["UserNumber"], 3)
		local userCmd = "030" .. userNumber .. "0"
		
		lastIndex = PadValue(ConvertToHex(4 + string.len(userCmd)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. userCmd)
		checkSum = GetCheckSum(request)

		LogDebug("ManageUserCodeCB Request(Shortcut): " .. request)
		SendCommand("\n" .. request .. checkSum)
		AddUserMenuShortcut = true

		return
	end

	if (UserCodeEnter ~= true) then
		local code = parms["UserCode"]

		lastIndex = PadValue(ConvertToHex(5 + string.len(code)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. code .. "B")
		checkSum = GetCheckSum(request)

		LogDebug("ManageUserCodeCB Request(Code): " .. request)
		SendCommand("\n" .. request .. checkSum)
		UserCodeEnter = true

		return
	end

	--*00#
	if (ExitAddUserCode ~= true) then

		local exitCmd = "A00B"
		lastIndex = PadValue(ConvertToHex(4 + string.len(exitCmd)))
		request = lastIndex .. "40" .. PadKeyPresses("10" .. exitCmd)
		checkSum = GetCheckSum(request)

		LogDebug("ManageUserCodeCB Request(Exit): " .. request)
		SendCommand("\n" .. request .. checkSum)
		ExitAddUserCode = true
	end

	--#
	lastIndex = PadValue(ConvertToHex(5))
	request = lastIndex .. "40" .. PadKeyPresses("10B")
	checkSum = GetCheckSum(request)

	LogDebug("ManageUserCodeCB Request(Done): " .. request)
	SendCommand("\n" .. request .. checkSum)

	UserCodeEnter = false
	ExitAddUserCode = false
	AddUserMenuShortcut = false

	KillTimer(ManageUserCodeTimer)
end