--[[=============================================================================
    Template code for a Security Panel/Partition serial driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "lib.c4_timer"
require "device_messages"
require "connections"
require "deocean_callbacks"
require "partition_proxy.securitypartition"

gPartitionIndex	= 1
gDELAYTIMER		= 0
--[[=============================================================================
    Interface routines required by the SecurityPanel code
===============================================================================]]
--[[=============================================================================
    Read information from the panel hardware about partitions and zones
===============================================================================]]
function SecCom_ReadPanelInfo()
	EstablishCommunication()
end

--[[=============================================================================
    Convert the given date/time parameters into the format required by the
    security panel

    Time Command: 9,(code),*020,<hhmm>,#
    Date Command: 9,(code),*021,<mmddyy>,#*00#
===============================================================================]]
function SecCom_SendDateAndTime(TargYear, TargMonth, TargDay, TargHour, TargMinute, TargSecond, InterfaceID, UserCode)

	if (UserCode == nil or UserCode == "") then
		local ParmList = TargYear .. "," .. TargMonth .. "," .. TargDay .. "," .. TargHour .. "," .. TargMinute .. "," .. TargSecond

		PartitionRequestAdditionalInfo(1, GetText("604D"), ParmList, "SecCom_SendDateAndTime", true, InterfaceID)
	else
		LogDebug("THERE")
		LogDebug("SendDateAndTime(): %s/%s/%s %s:%s", PadTimeDateValue(TargMonth), PadTimeDateValue(TargDay), PadTimeDateValue(TargYear), PadTimeDateValue(TargHour), PadTimeDateValue(TargMinute))
		local timeDate = { ["year"] = PadTimeDateValue(TargYear),
						   ["month"] = PadTimeDateValue(TargMonth),
						   ["day"] = PadTimeDateValue(TargDay),
						   ["hour"] = PadTimeDateValue(TargHour),
						   ["minute"] = PadTimeDateValue(TargMinute),
						   ["second"] = PadTimeDateValue(TargSecond),
						   ["code"] = UserCode }

		TimeAndDateTimer = CreateTimer("TimeShortcut", 500, "MILLISECONDS", SetDateAndTimeCB, true, timeDate)
		local lastIndex = PadValue(ConvertToHex(5 + string.len(UserCode)))
		local request = lastIndex .. "40" .. PadKeyPresses("109" .. UserCode)
		local checkSum = GetCheckSum(request)

		SendCommand("\n" .. request .. checkSum)
		StartTimer(TimeAndDateTimer)
	end
end

--[[=============================================================================
    Convert the given zone parameters into the format required by the panel
===============================================================================]]
function SecCom_SendSetZoneInfo(ZoneID, ZoneName, ZoneType, InterfaceID, UserCode)
	local entry
	local labels = {}
	local position
	local zoneText = ZoneName
	local exitLoop = false
	local zoneLabel = {}
	local stringCount = 1

	if (UserCode == nil or UserCode == "") then
		local ParmList = ZoneID .. "," .. ZoneName .. "," .. ZoneType

		PartitionRequestAdditionalInfo(1, GetText("604D"), ParmList, "SecCom_SendSetZoneInfo", true, InterfaceID)
	else
		-- Set the zone info to store the new type information if there is any
		SetZoneInfo(ZoneID, zoneText, GetZoneType(ZoneID), ZoneType)
		labels[0] = tonumber(ZoneID)

		while (stringCount <= 16 and exitLoop == false) do

			position = string.find(zoneText, " ")
			if (position ~= nil) then
				zoneLabel = string.sub(zoneText, 1, position-1)
				zoneText = string.sub(zoneText, position+1)
			else
				zoneLabel = zoneText
				exitLoop = true
			end

			entry = FindLabel(zoneLabel)
			if (entry ~= nil) then
				labels[stringCount] = entry
				stringCount = stringCount + 1
			else
				if (string.len(zoneLabel) > 0) then
					for i = 1, string.len(zoneLabel) do
						if (stringCount > 16) then
							break
						end

						labels[stringCount] = FindLabel(string.sub(zoneLabel, i, i))
						stringCount = stringCount + 1
					end

					if (stringCount < 16) then
						labels[stringCount] = FindLabel(" ")
						stringCount = stringCount + 1
					end
				else
					labels[stringCount] = FindLabel(" ")
					stringCount = stringCount + 1
				end
			end
		end

		if (stringCount > 16) then
			stringCount = stringCount - 1
		end

		labels["COUNT"] = stringCount

		ZoneLabelTimer = CreateTimer("ZoneLabelTimer", 1, "SECOND", ZoneLabelCB, true, labels)
		local lastIndex = PadValue(ConvertToHex(5 + string.len(UserCode)))
		local request = lastIndex .. "40" .. PadKeyPresses("108" .. UserCode)
		local checkSum = GetCheckSum(request)

		SendCommand("\n" .. request .. checkSum)
		StartTimer(ZoneLabelTimer)
	end
end

--[[=============================================================================
    Tell the panel to activate or de-activate a specific partition

    Note:  Not all panels will allow this
===============================================================================]]
function SecCom_SendPartitionEnabled(PartitionIndex, IsEnabled, InterfaceID)
	-- NOTE: This feature is not supported by the panel.
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--  Interface routines required by the SecurityParition code
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[=============================================================================
    Will issue a arm request to the panel

    ArmType                         Keypad Command Sequence
    ------------------------------------------------------------
    Stay                            2,(code)
    Away                            3,(code)
    Stay(No Delay)                  2,(code),4
    Away(No Delay)                  3,(code),4
    Stay(Latchkey)                  2,(code),6
    Away(LatchKey)                  3,(code),6
    Stay(Silent)                    5,2,(code)
    Away(Silent)                    5,3,(code)
    Stay(Lights Off)                2,(code),0
    Away(Lights Off)                3,(code),0
===============================================================================]]
function EntryDelay()
     SetPartitionState(gPartitionIndex,AS_ARMED,"Entry Delay",gDELAYTIMER,gDELAYTIMER)
	LogTrace("KillTimer: EntryDelayTimer .. %ds",gDELAYTIMER)  
	KillTimer(EntryDelayTimer)
end

function SecCom_SendArmPartition(PartitionIndex, ArmType, UserCode, Bypass, InterfaceID)
	local command
	local armType = "2"

	Bypass = Bypass or false
	UserCode = UserCode or ""
	LogTrace("SendArmPartition:   PartitionIndex[%s]   ArmType[%s]   UserCode[%s]   Bypass[%s]", PartitionIndex, ArmType, UserCode, tostring(Bypass))
	   
	
	gPartitionIndex = PartitionIndex
	   
	if ArmType == "立即布防" then
	   gDELAYTIMER = 1
	end
	if  ArmType == "延迟布防（30s）" then
	   gDELAYTIMER = 30
	end
	if  ArmType == "延迟布防（45s）" then
	   gDELAYTIMER = 45
	end
	if  ArmType == "延时布防（1min）" then
	   gDELAYTIMER = 60
	end
	if  ArmType == "延时布防（2min）" then
	   gDELAYTIMER = 120
	end
	if  ArmType == "延时布防（5min）" then
	   gDELAYTIMER = 300
	end
	if  ArmType == "延时布防（10min）" then
	   gDELAYTIMER = 600
	end
	
     SetPartitionState(PartitionIndex,AS_EXIT_DELAY,"Entry Delay",gDELAYTIMER,gDELAYTIMER)  ----
     EntryDelayTimer = CreateTimer("EntryDelayTimer", gDELAYTIMER, "SECONDS", EntryDelay, false)
     StartTimer(EntryDelayTimer)
     LogTrace("StartTimer: EntryDelayTimer .. %ds",gDELAYTIMER) 
	
	if (string.lower(tostring(Bypass)) ~= "true") then
		  
		ArmType = string.lower(tostring(ArmType))
		if (string.starts(ArmType, "stay") ~= true) then
			armType = "3"
		end

		if (PanelFeatures[PartitionIndex]["Quick Arm"] == 1) then
			armType = armType .. UserCode
		end
		
		if ((string.find(ArmType, "no delay")) ~= nil) then
			armType = armType .. "4"
		elseif ((string.find(ArmType, "latchkey")) ~= nil) then
			armType = armType .. "6"
		elseif ((string.find(ArmType, "silent")) ~= nil) then
			armType = "5" .. armType
		else
			if ((string.find(ArmType, "lights off")) ~= nil) then
				armType = armType .. "0"
			end
		end

		ClearInterfaceTimer = CreateTimer("ResetInterfaceID", 5, "SECONDS", ClearInterfaceID, false)
		LAST_INTERFACEID = InterfaceID
	else
		armType = "#"
		LAST_INTERFACEID = ""
		KillTimer(ClearInterfaceTimer)
	end

	command = GetCommandString(PartitionIndex, armType)
	SendCommand(command)

	if (ClearInterfaceTimer == nil) then
		StartTimer(ClearInterfaceTimer)
	end
end

--[[=============================================================================
    Will identify if a arm request has been canceled by the UI
===============================================================================]]
function SecCom_ArmCancel(PartitionIndex, InterfaceID)
	LogTrace("SendArmCancel:   PartitionIndex[%s]", PartitionIndex)

	command = GetCommandString(PartitionIndex, "*")
	SendCommand(command)
end

--[[=============================================================================
    Will issue a disarm request to the panel
===============================================================================]]
function SecCom_SendDisarmPartition(PartitionIndex, UserCode, InterfaceID)
	local command
	local disarmCmd = "1" .. UserCode

	LogTrace("SendDisarmPartition:   PartitionIndex[%s]   UserCode[%s]", PartitionIndex, UserCode)
	
	FromSerialCommand = "PartitionData"
     FromSerialData = "0" .. PartitionIndex .. "0101"        
     DispatchMessage(FromSerialCommand, FromSerialData)
	
	KillTimer(EntryDelayTimer)
	
	command = GetCommandString(PartitionIndex, disarmCmd)

	SendCommand(command)
end

--[[=============================================================================
    Will issue an emergency request to the panel

    EmergencyType
    ------------------------------------------------------------
    Fire
    Medical
    Police
    Panic
===============================================================================]]
function SecCom_SendExecuteEmergency(PartitionIndex, EmergencyType, InterfaceID)
	local command
	local emergencyCmd = ""

	LogTrace("SendExecuteEmergency:   PartitionIndex[%s]   EmergencyType[%s]", PartitionIndex, EmergencyType)
	if (EmergencyType == "Fire") then
		emergencyCmd = "4E"
	elseif (EmergencyType == "Medical") then
		emergencyCmd = "4D"
	elseif (EmergencyType == "Police") then
		emergencyCmd = "4C"
	elseif (EmergencyType == "Panic") then
		emergencyCmd = "4D"
	else
		return
	end

	local lastIndex = PadValue(ConvertToHex(5))
	local request = lastIndex .. "40" .. PadKeyPresses(PartitionIndex .. "0") .. emergencyCmd
	local checkSum = GetCheckSum(request)

	LogDebug("SendExecuteEmergency:   CommandString[%s]   CheckSum[%s]", request, checkSum)
	SendCommand("\n" .. request .. checkSum)
end

--[[=============================================================================
    Will issue a bypass for the indicated zone

    Directly Bypass:   #,(code),<zone#>
    Indirectly Bypass: 2,(code),# or 3,(code)
===============================================================================]]
function SecCom_SendBypassZoneCommand(PartitionIndex, ZoneID, DoBypass, UserCode, InterfaceID)
	local command

	LogTrace("SendBypassZoneCommand:   PartitionIndex[%d]   ZoneID[%d]   DoByPass[%s]   UserCode[%d]", tonumber(PartitionIndex), tonumber(ZoneID), tostring(DoBypass), tonumber(UserCode))

	-- See if partition is armed if so we are doing a direct bypass
	if (IsPartitionArmed(PartitionIndex)) then
		command = "#" .. UserCode .. ZoneID .. "#"
	else
		local armType = GetPartitionStateType(PartitionIndex)

		if (armType == "Stay") then
			command = "2" .. UserCode .. "#"
		else
			command = "3" .. UserCode .. "#"
		end
	end

	command = GetCommandString(tostring(PartitionIndex), command)
	SendCommand(command)
end

--[[==========================================================================================
    Sent to emulate key press commands from a keypad

    Format: [PN][AN][KP1] … [KPn]

    KeyCode             Value
    ------------------------------------------------------------
    00h                 0
    01h                 1
    02h                 2
    03h                 3
    04h                 4
    05h                 5
    06h                 6
    07h                 7
    08h                 8
    09h                 9
    0Ah                 *
    0Bh                 #
    0Ch                 Police Panic
    0Dh                 Aux. Panic
    0Eh                 Fire Panic
    10h                 Lights On
    11h                 Lights Off
    12h                 Lights Toggle
    13h                 Keyswitch on
    14h                 Keyswitch off
    15h                 Keyswitch toggle (not implemented)
    16h                 <undefined>
    17h                 <undefined>
    18h                 <undefined>
    19h                 <undefined>
    1Ah                 <undefined>
    1Bh                 <undefined>
    1Ch                 Fire TP – Acknowledge
    1Dh                 Fire TP – Silence
    1Eh                 Fire TP – Fire Test
    1Fh                 Fire TP – Smoke Reset
    20h                 Keyfob disarm
    21h                 Keyfob arm
    22h                 Keyfob lights
    23h                 Keyfob star
    24h                 Keyfob arm/disarm
    25h                 Keyfob lights/star
    26h                 Keyfob long lights
    27h                 Keyfob direct arm to level 3
    28h                 Keyfob direct arm to level 2
    29h                 Keyfob arm/star
    2Ah                 Keyfob disarm/lights
    2Ch                 TP A key
    30h                 TP B key
    2Dh                 TP C key
    33h                 TP D key
    2Eh                 TP E key
    36h                 TP F key

    Bit 6 = held for a few seconds

============================================================================================]]
function SecCom_SendKeyPress(PartitionIndex, KeyValue, InterfaceID)

	local lastIndex = PadValue(ConvertToHex(5))
	local request = lastIndex .. "40" .. PadKeyPresses(PartitionIndex .. "0") .. KeypadValues[KeyValue]
	local checkSum = GetCheckSum(request)

	LogDebug("CommandKeypress:   PartitionID[%d]   Key[%s]", PartitionIndex, KeyValue)
	SendCommand("\n" .. request .. checkSum)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--  The following functions are called from the PgmInfo code 
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[=============================================================================
    Send the open control command to the specified pgm
===============================================================================]]
function SecCom_SendPgmControlOpen(PgmID)

	LogDebug("SecCom_SendPgmControlOpen:")
	if (IsPgmOpen(PgmID) ~= true) then
		SecCom_SendPgmControlToggle(PgmID)
	end
end

--[[=============================================================================
    Send the closed control command to the specified pgm
===============================================================================]]
function SecCom_SendPgmControlClose(PgmID)

	LogDebug("SecCom_SendPgmControlClose:")
	if (IsPgmOpen(PgmID) == true) then
		SecCom_SendPgmControlToggle(PgmID)
	end
end

--[[=============================================================================
    Send the control command to toggle the current state of the specified pgm
===============================================================================]]
function SecCom_SendPgmControlToggle(PgmID)
	local lastIndex = PadValue(ConvertToHex(7))
	local request = lastIndex .. "40" .. PadKeyPresses("1077" .. tostring(PgmID))
	local checkSum = GetCheckSum(request)

	LogDebug("SecCom_SendPgmControlToggle:   Request[%s]", request)
	SendCommand("\n" .. request .. checkSum)
end

--[[=============================================================================
    Send the control command to send a pulse to the specified pgm
===============================================================================]]
function SecCom_SendPgmControlTrigger(PgmID, Time)
	local lastIndex = PadValue(ConvertToHex(7))
	local request = lastIndex .. "40" .. PadKeyPresses("1077" .. tostring(PgmID))
	local checkSum = GetCheckSum(request)
	local command = "\n" .. request .. checkSum

	LogDebug("SecCom_SendPgmControlTrigger:   Request[%s]", request)
	TriggerPgmTimer = CreateTimer("TriggerPgm", Time, "MILLISECONDS", TriggerPgmCB, false, command)

	SendCommand(command)
	StartTimer(TriggerPgmTimer)
end
