--[[=============================================================================
    Functions for Getting, Handling, and Dispatching Messages

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "lib.c4_timer"
require "lib.c4_object"
require "common.c4_serial_connection"
require "panel_proxy.securitypanel_functions"
require "partition_proxy.securitypartition"

require "panel_info"
require "panel_strings"
require "alarm_trouble"

ack  = string.char(0x06)
nak  = string.char(0x15)
fail = string.char(0xff)

language = "EN"
bufferPosition = 1
RealArmedType = ""
RealArmedState = ""

if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "1.0.0"
end

--[[=============================================================================
    Send the given command string to the panel
===============================================================================]]
function SendCommand(CommandStr)
	--gCon:SendCommand(CommandStr)
end

--[[=============================================================================
    Send an NAK to the panel
===============================================================================]]
function SendNak()
	-- LogDebug("Sending NAK")
	--SendCommand(nak)
end

--[[=============================================================================
    Send an ACK to the panel
===============================================================================]]
function SendAck()
	-- LogDebug("Sending ACK")
	--SendCommand(ack)
end

--[[==========================================================================================
    Take care of any initial setup when the serial connection is first established, and
    perform any handshaking or other similar start up code here.
============================================================================================]]
function EstablishCommunication()
	LANG = Properties['Language']

	SYSTEM_INITIALIZED = false
	CheckInitTimer = CreateTimer("CheckInitTimer", 30, "SECONDS", CheckInitCB, false)
	StartTimer(CheckInitTimer)

	DynamicDataRefresh()
end

--[[=============================================================================
    TearDownCommunication()

    Description:
    Called when the serial cable is disconnected from the panel.
===============================================================================]]
function TearDownCommunication()
	SYSTEM_INITIALIZED = false
end

--[[=============================================================================
    Get the next byte from the gReceiveBuffer and return it
===============================================================================]]
function GetHexByte()
	local nibble
	local reply = ""

	for _ = 1, 2 do
		nibble = string.sub(gReceiveBuffer, bufferPosition, bufferPosition)

		if (not((nibble >= "0" and nibble <= "9") or (nibble >= "A" and nibble <= "F"))) then
			-- LogDebug("GetHexByte(): Invalid character returned [%s]", tostring(nibble))
			gReceiveBuffer = string.gsub(gReceiveBuffer, nibble, "")

			return nil
		end

		reply = reply .. nibble
		bufferPosition = bufferPosition + 1
	end

	--LogDebug("GetHexByte(): reply[%s]", tostring(reply))

	return reply
end

--[[=============================================================================
    Validate that the given checksum is correct for the data
===============================================================================]]
function IsCheckSumValid(checkSum, data)
	local i = 1
	local current = 0
	local computedSum = 0
	local checkSumValid = false
	local length = string.len(data)

	-- LogDebug("IsCheckSumValid(): checkSum[%s],    Data[%s]", tostring(checkSum), tostring(data))
	while (i <= length) do

		current = string.sub(data, i, i + 1)
		computedSum =  computedSum + tonumber(string.format("0x%s", current))

		i = i + 2
	end

	computedSum = string.format("%02X", computedSum)
	computedSum = string.sub(computedSum, string.len(computedSum) - 1)
	-- LogDebug("IsCheckSumValid(): computedSum[%s]", computedSum)

	if computedSum == checkSum then
		checkSumValid = true
	end

	return checkSumValid
end

--[[=============================================================================
    Remove the characters upto position from the gReceiveBuffer
===============================================================================]]
function RemoveCharactersFromBuffer(position)
	gReceiveBuffer = string.sub(gReceiveBuffer, position+1)
	--LogDebug("RemoveCharactersFromBuffer(): Start Length[%d],    New Length[%d],    Position[%d]", startLength, string.len(gReceiveBuffer), position)
end

--[[=============================================================================
    Clears all the data from the gReceiveBuffer clearing the data from the
    panel from the message stream.
===============================================================================]]
function ClearMessageBuffer()
	gReceiveBuffer = ""
	LogTrace("ClearMessageBuffer(): ReceiveBuffer is now {{{%s}}}", gReceiveBuffer)
end

--[[=============================================================================
    GetMessage()
  
    Description:
    Used to retrieve a message from the communication buffer. Each driver is
    responsible for parsing that communication from the buffer.
  
    Parameters:
    None
  
    Returns:
    A single message from the communication buffer
===============================================================================]]
function GetMessage()
	local i = 1
	local count = 0
	local checkSum = 0
	local nextChar = ""
	local lastIndex = 0
	local comMessage = ""

	bufferPosition = 1
	while (string.len(gReceiveBuffer) > 0) do

		nextChar = string.format("%02X", string.byte(gReceiveBuffer, bufferPosition))
		-- LogDebug("GetMessage(): nextChar[%s]", tostring(nextChar))

		if (nextChar == string.format("%02X", string.byte('\n'))) then

			bufferPosition = bufferPosition + 1
			if (string.len(gReceiveBuffer) > 1) then

				-- Getting the index count of the remaining data
				lastIndex = GetHexByte()
				if (lastIndex == nil) then
					return nil
				end

				-- LogDebug("GetMessage(): lastIndex[%d]", tonumber(lastIndex, 16))

				count = tonumber(lastIndex, 16) - 1
				if (string.len(gReceiveBuffer) < count*2) then

					-- LogDebug("GetMessage(): We don't have the full string return until we have more in the buffer")
					return nil
				end

				-- Read the data values from the buffer
				while (i <= count) do

					nextChar = GetHexByte()
					if (nextChar == nil) then
						return nil
					end

					comMessage = comMessage .. nextChar
					i = i + 1
				end

				-- Checkout the checksum and make sure its good
				checkSum = GetHexByte()
				if (checkSum == nil) then
					return nil
				end

				if (not IsCheckSumValid(checkSum, lastIndex .. comMessage)) then
					-- LogDebug("GetMessage(): Checksum does not match the expected value")
					SendNak()

					return nil
				end

				RemoveCharactersFromBuffer(bufferPosition)
				SendAck()

				return comMessage
			else
				-- LogDebug("GetMessage(): We have a newline but nothing else in the buffer so lets wait for more of the command")
				return nil
			end
		else
			-- LogDebug("GetMessage(): Remove the junk from the serial port until we find a newline")
			RemoveCharactersFromBuffer(1)
		end
	end

	return comMessage
end

--[[=============================================================================
    HandleMessage(DataMessage)

    Description
    This is where we parse the messages returned from the GetMessage()
    function into a command and data. The call to 'DispatchMessage' will use the
    'name' variable as a key to determine which handler routine, function, should
    be called in the DEV_MSG table. The 'value' variable will then be passed as
    a string parameter to that routine.

    Parameters
    DataMessage(string) - Message string containing the function and value to
                          be sent to DispatchMessage

    Returns
    Nothing
===============================================================================]]
function HandleMessage(DataMessage)
	local command = nil
	local FromSerialData = ""
	local FromSerialCommand = ""

     LogDebug("HandleMessage(): DataMessage[%s]", DataMessage)

	command = string.sub(DataMessage, 1, 2)
	FromSerialData = string.sub(DataMessage, 3)

	if command == "01" then
		FromSerialCommand = "PanelType"
	elseif command == "02" then
		FromSerialCommand = "EventLost"
	elseif command == "03" then
		FromSerialCommand = "ZoneData"
	elseif command == "04" then
		FromSerialCommand = "PartitionData"
	elseif command == "05" then
		FromSerialCommand = "DeviceData"
	elseif command == "06" then
		FromSerialCommand = "DeviceCapabilities"
	elseif command == "07" then
		FromSerialCommand = "OutputData"
	elseif command == "08" then
		FromSerialCommand = "ListComplete"
	elseif command == "09" then
		FromSerialCommand = "UserData"
	elseif command == "0A" then
		FromSerialCommand = "ScheduleData"
	elseif command == "0B" then
		FromSerialCommand = "ScheduledEventData"
	elseif command == "0C" then
		FromSerialCommand = "LightSensorAttachment"
	elseif command == "20" then
		FromSerialCommand = "DynamicImage"
	elseif command == "21" then
		FromSerialCommand = "ZoneStatus"
	elseif command == "22" then
		FromSerialCommand, FromSerialData = ArmingCommands(FromSerialData)
	elseif command == "23" then
		FromSerialCommand, FromSerialData = ControlCommands(FromSerialData)
	else
		-- LogDebug("HandleMessage(): Unknown command returned from the panel")
		return nil
	end

	--[[if (FromSerialData == nil) then
		LogDebug("HandleMessage(): FromSerialCommand[%s],    FromSerialData[ ]", FromSerialCommand)
	else
		LogDebug("HandleMessage(): FromSerialCommand[%s],    FromSerialData[%s]", FromSerialCommand, FromSerialData)
	end]]

	DispatchMessage(FromSerialCommand, FromSerialData)
end

--[[=============================================================================
    ArmingCommands(DataMessage)

    Description
    This is a subset of the HandleMessage function that is used specifically to
    call the arm commands sent from the panel.

    Parameters
    DataMessage(string) - Message string containing the function and value to
                          be sent to DispatchMessage

    Returns
    The name of the DEV_MSG function to call as well as the data to be passed
    to that function
===============================================================================]]
function ArmingCommands(DataMessage)
	local command = nil
	local FromSerialData = ""
	local FromSerialCommand = ""

	--LogDebug("ArmingCommands(): DataMessage[%s]", DataMessage)

	command = string.sub(DataMessage, 1, 2)
	FromSerialData = string.sub(DataMessage, 3)

	if command == "01" then
		FromSerialCommand = "ArmingLevel"
	elseif command == "02" then
		FromSerialCommand = "AlarmTrouble"
	elseif command == "03" then
		FromSerialCommand = "EntryExitDelay"
	elseif command == "04" then
		FromSerialCommand = "SirenSetup"
	elseif command == "05" then
		FromSerialCommand = "SirenSynchronize"
	elseif command == "06" then
		FromSerialCommand = "SirenGo"
	elseif command == "09" then
		FromSerialCommand = "TouchpadDisplay"
	elseif command == "0B" then
		FromSerialCommand = "SirenStop"
	elseif command == "0C" then
		FromSerialCommand = "FeatureState"
	elseif command == "0D" then
		FromSerialCommand = "Temperature"
	elseif command == "0E" then
		FromSerialCommand = "TimeDate"
	else
		-- LogDebug("ArmingCommands(): Unknown secondary command returned from the panel")
		return nil
	end

	return FromSerialCommand, FromSerialData
end

--[[=============================================================================
    ControlCommands(DataMessage)

    Description
    This is a subset of the HandleMessage function that is used specifically to
    call the control commands sent from the panel.

    Parameters
    DataMessage(string) - Message string containing the function and value to
                          be sent to DispatchMessage

    Returns
    The name of the DEV_MSG function to call as well as the data to be passed
    to that function
===============================================================================]]
function ControlCommands(DataMessage)
	local command = nil
	local FromSerialData = ""
	local FromSerialCommand = ""

	--LogDebug("ControlCommands(): DataMessage[%s]", DataMessage)

	command = string.sub(DataMessage, 1, 2)
	FromSerialData = string.sub(DataMessage, 3)

	if command == "01" then
		FromSerialCommand = "LightState"
	elseif command == "02" then
		FromSerialCommand = "UserLights"
	elseif command == "03" then
		FromSerialCommand = "Keyfob"
	else
		-- LogDebug("ControlCommands(): Unknown secondary command returned from the panel")
		return nil
	end

	return FromSerialCommand, FromSerialData
end

--[[=============================================================================
    DispatchMessage(MsgKey, MsgData)

    Description
    Parse routine that will call the routines to handle the information returned
    by the connected system.

    Parameters
    MsgKey(string)  - The function to be called from within DispatchMessage
    MsgData(string) - The parameters to be passed to the function found in MsgKey

    Returns
    Nothing
===============================================================================]]
function DispatchMessage(MsgKey, MsgData)
	MsgData = MsgData or ""

	if (DEV_MSG[MsgKey] ~= nil and (type(DEV_MSG[MsgKey]) == "function")) then

		-- LogDebug("DEV_MSG.%s:  %s", tostring(MsgKey), tostring(MsgData))
		local status, err = pcall(DEV_MSG[MsgKey], MsgData)
		if (not status) then
			LogError("LUA_ERROR[DispatchMessage]: %s", tostring(err))
		end
	else
		LogTrace("HandleMessage: Unhandled command = %s", tostring(MsgKey))
	end
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Function from Automation to Panel
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[=============================================================================
    This command is sent to get the following lists of data
    - (03h)Zone Data
    - (04h)Partition Data
    - (05h)SuperBus Device Data
    - (06h)SuperBus Device Capabilities Data
    - (07h)Output Data
    - (09h)User Data
    - (0Ah)Schedule Data
    - (0Bh)Scheduled Event Data
    - (0Ch)Light to Sensor Attachment
===============================================================================]]
function FullEquipmentList()
	local command = string.char(0x0a, 0x30, 0x32, 0x30, 0x32, 0x30, 0x34)

	-- SYSTEM_INITIALIZED = false
	LogTrace("FullEquipmentList(): command[02h 02h 04h]")
	SendCommand(command)
end

--[[=============================================================================
    This command is sent to get one of the following lists of data
    - (03h)Zone Data
    - (04h)Partition Data
    - (05h)SuperBus Device Data
    - (06h)SuperBus Device Capabilities Data
    - (07h)Output Data
    - (09h)User Data
    - (0Ah)Schedule Data
    - (0Bh)Scheduled Event Data
    - (0Ch)Light to Sensor Attachment
===============================================================================]]
function SingleEquipmentList(listType)
	local command = "03020" .. listType
	local checkSum = GetCheckSum(command)

	LogTrace("SingleEquipmentList(): command[03h 02h 0%sh %sh]", listType, checkSum)
	SendCommand("\n" .. command .. checkSum)
end

--[[=============================================================================
    Sent to request data refreshing. The Panel sends Arming Level, non-normal Zone Status,
    Alarm/Trouble Status commands in response.

    Concord will also respond with the Panel Type(01h), Feature State(22h/0Ch),
    Temperature(22h/0Dh), Light State(23h/01h), and Time and Date(22h/0Eh) commands.
===============================================================================]]
function DynamicDataRefresh()
	local command = "\n022022"

	LogTrace("DynamicDataRefresh(): command[02h 20h 22h]")
	StartupTimer = CreateTimer("StartupTimer", 10, "SECONDS", StartupCB, false)
	StartTimer(StartupTimer)
	
	SendCommand(command)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Routines to handle messages from the serial port
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[=============================================================================
    Command is sent on panel power up and when a communication failure restoral with the
    Superbus2000 occurs.
===============================================================================]]
function DEV_MSG.PanelType(reply)
	local buffer = { reply }

	LogTrace("DEV_MSG.PanelType(): reply[%s]", reply)
	SetPanelType(PullValue(buffer))

	local hwRevision
	local hwRevisionHigh = PullValue(buffer)

	if (IsConcord()) then
		hwRevisionHigh = hwRevisionHigh + 64
		hwRevision = string.char(hwRevisionHigh)
	else
		hwRevision = hwRevisionHigh
	end

	hwRevision = hwRevision .. PullIntValue(buffer)
	SetHWRevision(hwRevision)

	-- NOTE: Really need to figure out what the SW revision is all about right now we'll just truncate it together
	local swRevision = PullValue(buffer) .. PullValue(buffer)
	SetSWRevision(swRevision)

	local serialNumber = "00000000"
	if (IsConcord()) then
		serialNumber = ""
		for _ = 1, 4 do
			serialNumber = serialNumber .. PullValue(buffer)
		end
	end
	SetSerialNumber(serialNumber)

	LogDebug("DEV_MSG.PanelType(): Brand[%s],    Type[%s],    HWRevision[%s],    SWRevision[%s],    SerialNumber[%s]", GetBrand(), GetPanelType(), GetHWRevision(), GetSWRevision(), GetSerialNumber())
end

--[[=============================================================================
    Sent when buffer has overflowed resulting in the loss of system events.
===============================================================================]]
function DEV_MSG.EventLost(reply)
	LogTrace("DEV_MSG.EventLost()")

	DynamicDataRefresh()
	RecordWarningEvent("System", "Buffer Overflow", "Panel Event Lost Issuing Refresh and List Commands")
end

--[[=============================================================================
    Sent in response to Full Equipment List(02h) or Single Equipment List(02h/03h) request.
    This will be sent for each zone that has been programmed.

    Format: [PN][AN][GN][ZNh][ZNl][ZT][ZS][text]

    GroupNumber
    ------------------------------------------------------------
    0-35 - This is what most people would call the zone type
    
    ZoneType
    ------------------------------------------------------------
    0 = harwired
    1 = RF
    2 = RF Touchpad

    ZoneState
    ------------------------------------------------------------
    bit 0: 1 = opened;       zone opened or closed
    bit 1: 1 = fault;        unused will always be 0
    bit 2: 1 = alarm;        opened when armed
    bit 3: 1 = trouble;      problem with the zone should write to trouble text
    bit 4: 1 = bypassed;     zone has been bypassed
===============================================================================]]
function DEV_MSG.ZoneData(reply)
	local buffer = { reply }

	LogTrace("DEV_MSG.ZoneData(): reply[%s]", reply)

	local partitionID = PullIntValue(buffer)
	local areaNumber = PullIntValue(buffer)
	local groupNumber = PullIntValue(buffer)

	-- Unused value remove from buffer
	PullValue(buffer)

	local zoneID = partitionID*100+50+PullIntValue(buffer)
	local zoneType = PullIntValue(buffer)
	local zoneState = PullIntValue(buffer)

	local zoneText = ""
	local zoneTokens = string.sub(buffer[1], 1)
	if (string.len(zoneTokens) > 0) then
		zoneText = GetText(zoneTokens)
	else
		--zoneText = sensor .. " " .. PadValue(zoneID)
		zoneText = "Partition " .. tostring(partitionID) .. " Sensor" .. tostring(zoneID%100-50)--百位数用于区分是哪个防区
	end

	local zoneLabel = ""
	if (string.len(zoneText) > 0) then
		zoneLabel = "Label[" .. zoneText .. "]"
	end

	LogTrace("DEV_MSG.ZoneData(): Partition[%d]   Zone(ID[%d], Type[%d], State[%d]) %s", partitionID, zoneID, groupNumber, zoneState, zoneLabel)

	local c4ZoneType = GetC4ZoneType(groupNumber)
	SetZoneInfo(zoneID, zoneText, groupNumber, c4ZoneType)

	AddZoneToPartition(partitionID, zoneID)
	UpdateZoneState(zoneID, zoneState)
end

--[[=============================================================================
    Sent in response to Full Equipment List(02h) or Single Equipment List(02h/04h) request.
    This will be sent for each enabled partition.

    Format: [PN][AN][AL][text]

    Arming Level
    ------------------------------------------------------------
    1 = Off
    2 = Stay
    3 = Away
    8 = Phone Test
    9 = Sensor Test
===============================================================================]]
function DEV_MSG.PartitionData(reply)
	local buffer = { reply }

	LogTrace("DEV_MSG.PartitionData(): reply[%s]", reply)

	local partitionID = PullIntValue(buffer)

	if (IsPartitionEnabled(partitionID)) then
		local areaNumber = PullIntValue(buffer)
		local armLevel = PullIntValue(buffer)

		local partitionText = ""
		local partitionTokens = string.sub(buffer[1], 1)
		if (string.len(partitionTokens) > 0) then
			partitionText = GetText(partitionTokens)
		end

		local partitionLabel = ""
		if (string.len(partitionText) > 0) then
			partitionLabel = "    Label[" .. partitionText .. "]"
		end

		LogTrace("DEV_MSG.PartitionData(): Partition[%d]    ArmLevel[%d]%s", partitionID, armLevel, partitionLabel)
		UpdatePartitionStatus(partitionID, armLevel)
	else
		LogTrace("DEV_MSG.PartitionData(): Partition[%d] marked as disabled", partitionID)
	end
end

--[[=============================================================================
    Sent for each enrolled bus device, in response to an equipment list request.

    Format: [PN][AN][ID1][ID2][ID3][DS][UN][text]
===============================================================================]]
function DEV_MSG.DeviceData(reply)
	local buffer = { reply }

	LogTrace("DEV_MSG.DeviceData(): reply[%s]", reply)
	local partitionID = PullIntValue(buffer)

	-- Area Number not used
	PullValue(buffer)

	local deviceIDHigh = PullIntValue(buffer)
	local deviceIDMid = PullIntValue(buffer)
	local deviceIDLow = PullIntValue(buffer)

	local deviceID = deviceIDHigh .. deviceIDMid .. deviceIDLow
	local deviceStatus = PullIntValue(buffer)

	if (deviceStatus == 1) then
		RecordWarningEvent("System", "Device", "Device[" .. deviceID .. "] failed")
	end
end

--[[=============================================================================
    Sent for each capability for each enrolled bus device, in response to an equipment list
    request. For each capability present there will be one byte specifying a capability number
    and possibly one or more bytes of capability dependent data that consists of the elabora-
    tion of the capability.

    Format: [ID1][ID2][ID3][CN][CD]

    Capability Number       Description                         Optional Data
    ------------------------------------------------------------------------------------
    0x00                    Power Supervision
    0x01                    Access Control
    0x02                    Analog Smoke
    0x03                    Audio Listen-In
    0x04                    SnapCard Supervision
    0x05                    Microburst
    0x06                    Dual Phone Line
    0x07                    Energy Management
    0x08                    Input Zones                         Number of inputs
    0x09                    Phast/Automation/System Manager
    0x0A                    Phone Interface
    0x0B                    Relay Outputs                       Number of outputs
    0x0C                    RF Receiver
    0x0D                    RF Transmitter
    0x0E                    Parallel Printer
    0x10                    LED Touchpad
    0x11                    1-Line/2-Line/BLT Touchpad
    0x12                    GUI Touchpad
    0x13                    Voice Evacuation
    0x14                    Pager
    0x15                    Downloadable code/data
    0x16                    JTECH Premise Pager
    0x17                    Cryptography
    0x18                    LED Display
===============================================================================]]
function DEV_MSG.DeviceCapabilities(reply)
	LogTrace("DEV_MSG.DeviceCapabilities(): reply[%s]", reply)
	-- NOTE: This is not something we are interested in tracking for the driver
end

--[[=============================================================================
    Sent for each programmed output in response to an equipment list request.

    Format: [ONh][ONl][OS][ID1][ID2][ID3][ID4][ID5][text]
===============================================================================]]
function DEV_MSG.OutputData(reply)
	local buffer = { reply }

	LogTrace("DEV_MSG.OutputData(): reply[%s]", reply)
	PullIntValue(buffer)
	local outputNumber  = PullIntValue(buffer)

	local outputState = PullValue(buffer)
	LogTrace("DEV_MSG.OutputData(): OutputStateFlag[%s]", string.sub(outputState, 1, 1))
	LogTrace("DEV_MSG.OutputData(): OutputStateFlag[%s]", string.sub(outputState, 2, 2))

	local isOpen = false
	if (string.sub(outputState, 2, 2) == "1") then
		isOpen = true
	end

	local ID1 = PullIntValue(buffer)
	local ID2 = PullIntValue(buffer)
	local ID3 = PullIntValue(buffer)
	local ID4 = PullIntValue(buffer)
	local ID5 = PullIntValue(buffer)

	local stringOutput = ""
	local textOutput = PullValue(buffer)
	while (textOutput ~= nil) do
		if (textOutput == "") then
			break
		end

		stringOutput = stringOutput .. FindWord(textOutput)
		textOutput = PullValue(buffer)
	end

	outputNumber = (tonumber(outputNumber) - 64)
	if (ID1 == 0 and ID2 == 0) then

		if (ID4 == 1) then
			LogTrace("DEV_MSG.OutputData():    Snapcard Output[%d]    State[%s]    Output[%s]", outputNumber, tostring(isOpen), tostring(stringOutput))
		else
			LogTrace("DEV_MSG.OutputData():    Onboard Output[%d]    State[%s]    Output[%s]", outputNumber, tostring(isOpen), tostring(stringOutput))
		end
	end

	local isControllable = Properties["Output " .. outputNumber .. " Controllable"]
	if (toboolean(isControllable)) then
		if (IsPgmValid(outputNumber) == false) then
			AddPgm(outputNumber)
		end

		SetPgmState(outputNumber, isOpen, false)
	end
end

--[[=============================================================================
    Sent after all the 03h and 05h commands have been sent in response to an
    equipment list request.
===============================================================================]]
function DEV_MSG.ListComplete(reply)

	LogTrace("DEV_MSG.ListComplete()")
	for i = 1, PARTITION_ID_MAX do
		if (GetPartitionZoneCount(i) == 0) then
			SetPartitionEnabled(i, false)
			C4:DeleteVariable("PARTITION" .. i .. "_LAST_USER")
			C4:DeleteVariable("PARTITION" .. i .. "_LAST_EVENT")
		else
			SetPartitionEnabled(i, true)
			C4:AddVariable("PARTITION" .. i .. "_LAST_USER", "-1", "NUMBER", true)
			C4:AddVariable("PARTITION" .. i .. "_LAST_EVENT", "", "STRING", true)
		end
		SetPartitionInitializingFlag(i, false)
	end

	SYSTEM_INITIALIZED = true
end

--[[=============================================================================
    Sent in response to Full Equipment List(02h) or Single Equipment List(02h/09h) request.
    This will be sent for each enabled partition.

    Format:
        [UNh][UNl][text]
            or
        [UNh][UNl][UCh][UCm][UCl][text]
===============================================================================]]
function DEV_MSG.UserData(reply)
	local buffer = { reply }

	LogTrace("DEV_MSG.UserData(): reply[%s]", reply)

	local userNumber = PullValue(buffer) .. PullValue(buffer)
	userNumber = tonumber(userNumber, 16)

	-- Code High value is unused
	PullValue(buffer)
	local code = PullValue(buffer) .. PullValue(buffer)

	-- NOTE: Not sure we need to do anything from a control4 perspective with user codes,
	--       we'll just leave the code for a while until we decide?

	-- LogTrace("DEV_MSG.UserData(): UserNumber[%s]: %s", userNumber, code)
end

--[[=============================================================================
    Sent for each schedule in response to an equipment list request.

    Format: [PA][AA][SN][SH][SM][PH][PM][DY][SV]
===============================================================================]]
function DEV_MSG.ScheduleData(reply)
	LogTrace("DEV_MSG.ScheduleData(): reply[%s]", reply)
	-- NOTE: This is not something we are interested in tracking for the driver
end

--[[=============================================================================
    Sent for each scheduled event in response to an equipment list request.

    Note: If an event is not attached to any schedules, it will not be sent in
    this list.

    Format: [PN][AN][ET][S1][S2][S3][S4]
===============================================================================]]
function DEV_MSG.ScheduledEventData(reply)
	local buffer = { reply }
	local description

	LogTrace("DEV_MSG.ScheduledEventData(): reply[%s]", reply)

	local partitionID = PullIntValue(buffer)
	local areaNumber = PullIntValue(buffer)
	local eventType = PullIntValue(buffer)

	if (eventType >= 1 and eventType <= 9) then
		description = string.format("Light %d switched", eventType)
	elseif (eventType >= 113 and eventType <= 118) then
		eventType = eventType - 112
		description = string.format("Output %d activated", eventType)
	elseif (eventType == 224) then
		description = "Latchkey Opened"
	elseif (eventType == 225) then
		description = "Latchkey Closed"
	elseif (eventType == 226) then
		description = "Exception Opened"
	elseif (eventType == 227) then
		description = "Exception Closed"
	elseif (eventType == 228) then
		description = "Auto Arm(level 3)"
	else
		return
	end

	RecordInfoEvent("System", "Event", description)
end

--[[=============================================================================
    Sent for each light in response to an equipment list request.

    Format: [PN][AN][L1] … [Ln]
===============================================================================]]
function DEV_MSG.LightSensorAttachment(reply)
	-- LogTrace("DEV_MSG.LightSensorAttachment(): reply[%s]", reply)
	-- NOTE: This is not something we are interested in tracking for the driver
end

--[[=============================================================================
    Sent on panel power up initialization and when a communication failure
    restoral with the Automation Module occurs. The Concord also sends this
    command when user or installer programming mode is exited.
===============================================================================]]
function DEV_MSG.DynamicImage(reply)
	LogTrace("DEV_MSG.DynamicImage(): reply[]")

	-- Clear the message buffer so we start fresh
	ClearMessageBuffer()
	RecordInfoEvent("System", "Status", "A panel power up initialization or a communication failure restoral has happened")
end

--[[=============================================================================
    Sent when there is a change in zone state(e.g. trip, restore, alarm, cancel,
    trouble, restoral, bypass, unbypass).  Also, if the Automation Module
    requests a Dynamic Data Refresh Request(20h) this command will be sent for
    each zone that is not normal (i.e. any zone that is open (non restored), in
    alarm, troubled or bypassed).

    Format: [PN][AN][ZNh][ZNl][ZS]

    ZoneState
    ------------------------------------------------------------
       bit 0: 1 = tripped;      opened when not armed(think this is a motion detector)
       bit 1: 1 = fault;        opened when not armed
       bit 2: 1 = alarm;        opened when armed
       bit 3: 1 = trouble;      problem with the zone should write to trouble text
       bit 4: 1 = bypassed;     zone has been bypassed
===============================================================================]]
function DEV_MSG.ZoneStatus(reply)
	local buffer = { reply }

	LogTrace("DEV_MSG.ZoneStatus(): reply[%s]", reply)

	local partitionID = PullIntValue(buffer)
	local areaNumber = PullIntValue(buffer)
	local zoneHighNumber = PullIntValue(buffer)
	local zoneID = partitionID*100+PullIntValue(buffer)
	local zoneState = PullIntValue(buffer)
	   
	LogDebug("DEV_MSG.ZoneStatus(): Partition[%d],   AreaNumber[%d],   ZoneNumberHigh[%d],   ZoneNumberLow[%d],   ZoneState[%d]", partitionID, areaNumber, zoneHighNumber, zoneID, zoneState)
	
	if (IsZoneValid(zoneID) == false) then
		SingleEquipmentList(3)
	else
		UpdateZoneState(zoneID, zoneState)
	end
	
	if (zoneState%2) == 0 then
	   C4:SendToProxy(tonumber(partitionID*100 + 50 + zoneID), "CLOSED", {}, "NOTIFY")
	   LogDebug("OutZoneStatus(): Partition[%d],   ZoneNumber[%d],   ZoneState[CLOSED]", partitionID, zoneID)
     elseif (zoneState%2) == 1 then 	
	   C4:SendToProxy(tonumber(partitionID*100 + 50 + zoneID), "OPENED", {}, "NOTIFY")
	   LogDebug("OutZoneStatus(): Partition[%d],   ZoneNumber[%d],   ZoneState[OPENED]", partitionID, zoneID)
     end
end

--[[=============================================================================
    Sent whenever there is a change in the arming level.  Also, if the
    Automation Module requests a Dynamic Data Refresh Request(20h) this command
    will be sent for each partition that is enabled.

    Format: [PN][AN][UNh][UNl][AL]

    User Number Description         Concord
    ------------------------------------------------------------
    Regular User Codes              00-229
    Partition n Master Code         230-237
    Partition n Duress Code         238-245
    System Master Code              246
    Installer Code                  247
    Dealer Code                     248
    AVM Code                        249
    Quick Arm                       250
    Key Switch Arm                  251
    System                          252

    ArmLevel
    ------------------------------------------------------------
    0 = Zone Test
    1 = Off
    2 = Home/Perimeter
    3 = Away/Full
    4 = Night
    5 = Silent
===============================================================================]]
function DEV_MSG.ArmingLevel(command)
	local buffer = { command }

	LogTrace("DEV_MSG.ArmingLevel(): reply[%s]", command)

	local partitionID = PullIntValue(buffer)
	local areaNumber = PullIntValue(buffer)
	local userNumberHigh = PullIntValue(buffer)
	local userNumberLow = PullIntValue(buffer)
	local armLevel = PullIntValue(buffer)

	if (userNumberHigh == 0) then
		UpdateC4Variable(partitionID, "User" .. tostring(userNumberLow), "Arming Level " .. tostring(userNumberLow))
	else
		UpdateC4Variable(partitionID, "Keyfob" .. tostring(userNumberLow), "Arming Level " .. tostring(userNumberLow))
	end

	LogTrace("DEV_MSG.ArmingLevel(): Partition[%d],   ArmingLevel[%d]", partitionID, armLevel)
	UpdatePartitionStatus(partitionID, armLevel)
end

--[[=============================================================================
    This command is sent to identify alarm and trouble conditions as well as several other
    system events.  Events are specified by three numbers; General Type, Specific Type, and
    Event Specific Data.

    Format: [PN][AN][ST][SNh][SNm][SNl][GT][ST][ESh][ESl]

    SourceType
    ------------------------------------------------------------
    0 = bus device
    1 = local phone
    2 = zone
    3 = system
    4 = remote phone
===============================================================================]]
function DEV_MSG.AlarmTrouble(command)
	local buffer = { command }

	LogTrace("DEV_MSG.AlarmTrouble(): reply[%s]", command)

	local partitionID = PullIntValue(buffer)
	local areaNumber = PullIntValue(buffer)
	local srcType = PullIntValue(buffer)
	local srcNumHigh = PullValue(buffer)
	local srcNumMid = PullValue(buffer)
	local srcNumLow = PullValue(buffer)
	local generalType = PullIntValue(buffer)
	local specificType = PullIntValue(buffer)
	local eventHigh = PullValue(buffer) or ""
	local eventLow = PullValue(buffer) or ""

	local srcNumber
	if (srcType == 2) then
		srcNumber = srcNumMid .. srcNumLow
	else
		srcNumber = srcNumHigh .. srcNumMid .. srcNumLow
	end

	ProcessEvent(partitionID, srcType, srcNumber, generalType, specificType, eventHigh, eventLow)
end

--[[=============================================================================
    Sent whenever an entry or exit delay is started or ended.

    Format: [PN][AN][DF][DTh][DTl]

    Delay Flags
    ------------------------------------------------------------
    bit 5, 4:       00 = standard
                    01 = extended
                    10 = twice extended
    bit 6:          1 = exit delay
                    0 = entry delay
    bit 7:          1 = end delay
                    0 = start delay
===============================================================================]]
function DEV_MSG.EntryExitDelay(command)
	local buffer = { command }

	LogTrace("DEV_MSG.EntryExitDelay(): reply[%s]", command)

	local partitionID = PullIntValue(buffer)

	-- See if we have an exit entry delay zone in the partition
	local zoneIDs = GetPartitionZoneIDs(partitionID)
	local zoneCount = GetPartitionZoneCount(partitionID)

	local zoneType
	local haveStandard = false
	local haveExtended = false
	local haveTwiceExt = false

	for i = 1, zoneCount do
		zoneType = GetZoneType(zoneIDs[i])
		if (zoneType == 10) then
			haveStandard = true
		elseif (zoneType == 11) then
			haveExtended = true
		elseif (zoneType == 12) then
			haveTwiceExt = true
		end
	end

	if (haveStandard == false and haveExtended == false and haveTwiceExt == false) then
		LogDebug("DEV_MSG.EntryExitDelay(): No Exit/Entry Delay zones in this partition")
		return
	end

	-- Don't care about the area group so throw it out
	PullIntValue(buffer)

	local delayType
	local delayExt = "Standard"

	-- The type of delay we are receiving
	local flags = GetBits(PullValue(buffer))
	for key,value in pairs(flags) do
		LogTrace("DEV_MSG.EntryExitDelay(): Flag[%d]", tonumber(value))
	end

	if (flags[5] == 1) then
		delayExt = "Extended"
	elseif (flags[6] == 1) then
		delayExt = "Twice Extended"
	end
	LogTrace("DEV_MSG.EntryExitDelay(): DelayExtenstion[%s]", delayExt)

	local partitionState
	if (flags[7] == 1) then
		delayType = "Exit Delay"
		partitionState = AS_EXIT_DELAY
	else
		delayType = "Entry Delay"
		partitionState = AS_ENTRY_DELAY
	end
	LogTrace("DEV_MSG.EntryExitDelay(): DelayType[%s]", delayType)

	local endDelay = false
	if (flags[8] == 1) then
		endDelay = true
	end

	-- Get the delay time for the entry or exit delay combine the upper and lower time
	local timeHigh = PullIntValue(buffer)
	local timeLow = PullIntValue(buffer)
	local delayTime = tonumber(string.format("%d%d", timeHigh, timeLow))

	LogTrace("DEV_MSG.EntryExitDelay(): HaveStandard[%s]   HaveExtended[%s]   HaveTwiceExt[%s]", tostring(haveStandard), tostring(haveExtended), tostring(haveTwiceExt))

	-- Do we even have an entry/exit delay zone type in this partition if not skip the delay
	if (((haveStandard == true) and (delayExt == "Standard")) or
		((haveExtended == true) and ((delayExt == "Extended") or (delayExt == "Standard"))) or 
		((haveTwiceExt == true) and ((delayExt == "Twice Extended") or (delayExt == "Extended") or (delayExt == "Standard")))) then

		local currentArmedState = GetPartitionState(partitionID)
		local currentArmedType = GetPartitionStateType(partitionID)

		if (endDelay == true) then

			if (delayType == "Entry Delay") then
				if (currentArmedState ~= "DISARMED_READY") then
					SetPartitionState(tonumber(partitionID), RealArmedState, RealArmedType)
				else
					SetPartitionState(tonumber(partitionID), currentArmedState, currentArmedType)
				end
			else
				if (currentArmedState ~= "ALARM" or currentArmedState ~= "DISARMED_READY") then
					SetPartitionState(tonumber(partitionID), RealArmedState, RealArmedType)
				else
					SetPartitionState(tonumber(partitionID), currentArmedState, currentArmedType)
				end
			end

			RealArmedType = ""
			RealArmedState = ""
		else
			if (delayTime > 0) then

				if (not(IsPartitionInDelay(partitionID))) then
					RealArmedType = currentArmedType
					RealArmedState = currentArmedState
				end

				SetPartitionState(tonumber(partitionID), partitionState, delayType, delayTime, delayTime)
			end
		end

		LogTrace("DEV_MSG.EntryExitDelay(): Partition[%d]    Delay(Time[%d], Type[%s], Flag[%s], End[%s])", partitionID, delayTime, delayType, delayExt, tostring(endDelay))
		RecordInfoEvent("Partition", "Entry/Exit Delay", "Partition " .. partitionID .. " is in " .. delayType)
	else
		LogTrace("DEV_MSG.EntryExitDelay(): No Entry/Exit DelayExtension[%s]", delayExt)
	end
end

--[[=============================================================================
    Sent whenever there is something to output with the interior siren output, status beeps,
    siren sounds, etc. There are 4 cadence bytes. Each bit specifies a 125 mSec time slice,
    1=on, 0=off, beginning with the most significant bit of the first cadence byte. The
    combined cadence is repeated RP times (if RP = 0, the cadence is repeated continuously).

    Format: [PN][AN][RP][CD1][CD2][CD3][CD4]
===============================================================================]]
function DEV_MSG.SirenSetup(command)
	-- LogTrace("DEV_MSG.SirenSetup(): reply[%s]", command)
	-- NOTE: We don't currently support sound so we are not going to handle this right now
end

--[[=============================================================================
    Sent to synchronize all continuous cadences previously set up by the Siren Setup command.
===============================================================================]]
function DEV_MSG.SirenSynchronize(command)
	-- LogTrace("DEV_MSG.SirenSynchronize(): reply[%s]", command)
	-- NOTE: We don't currently support sound so we are not going to handle this right now
end

--[[=============================================================================
    Sent to begin all non-continuous cadences previously set up by the Siren Setup command.
===============================================================================]]
function DEV_MSG.SirenGo(command)
	-- LogTrace("DEV_MSG.SirenGo(): reply[%s]", command)
	-- NOTE: We don't currently support sound so we are not going to handle this right now
end

--[[=============================================================================
    Sends the touchpad display text tokens to the Automation Module.

    Format: [PN][AN][MT][DT]
===============================================================================]]
function DEV_MSG.TouchpadDisplay(command)
	local buffer = { command }

	LogTrace("DEV_MSG.TouchpadDisplay(): reply[%s]", command)

	local partitionID = PullIntValue(buffer)
	if (IsPartitionEnabled(partitionID)) then
		local areaNumber = PullIntValue(buffer)
		local messageType = PullIntValue(buffer)

		local typeLabel = "Broadcast"
		if (messageType == 0) then
			typeLabel = "Normal"
		end

		local displayText = ""
		local displayTokens = string.sub(buffer[1], 1)
		if (string.len(displayTokens) > 0) then
			displayText = GetText(displayTokens)
		end

		LogTrace("DEV_MSG.TouchpadDisplay(): Partition[%d]   Type[%s]   Message[%s]", partitionID, messageType, displayText)
		if (displayText ~= nil) then
			DisplayPartitionText(partitionID, displayText)
		end
	end
end

--[[=============================================================================
    Sent to stop any cadence being output.

    Format: [PN][AN]
===============================================================================]]
function DEV_MSG.SirenStop(command)
	-- LogTrace("DEV_MSG.SirenStop(): reply[%s]", command)
	-- NOTE: We don't currently support sound so we are not going to handle this right now
end

--[[=============================================================================
    Sent when a change in feature state occurs and in response to a 
    DynamicDataRefreshRequest.

    Format: [PN][AN][FS1]

    Feature State
    ------------------------------------------------------------
    Bit 0 = chime
    Bit 1 = energy saver
    Bit 2 = no delay
    Bit 3 = latchkey
    Bit 4 = silent arming
    Bit 5 = quick arm
    Bit set = on
    Bit clear = off
===============================================================================]]
function DEV_MSG.FeatureState(command)
	local buffer = { command }

	LogTrace("DEV_MSG.FeatureState(): reply[%s]", command)
	local partitionID = PullIntValue(buffer)

	-- area group not used
	PullValue(buffer)

	local featureState = PullIntValue(buffer)
	local featureBytes = GetBits(featureState)

	local state = ""
	local feature = ""
	local description = ""
	for i=1, 6 do
		description = " has been "

		if (i == 1) then
			feature = "Chime"
		elseif (i == 2) then
			feature = "Energy Saver"
		elseif (i == 3) then
			feature = "No Delay"
		elseif (i == 4) then
			feature = "Latchkey"
		elseif (i == 5) then
			feature = "Silent Arming"
		elseif (i == 6) then
			feature = "Quick Arm"
		end

		if (featureBytes[i] ~= nil) then
			state = GetFeatureState(featureBytes[i])
			PanelFeatures[partitionID][tostring(feature)] = tonumber(featureBytes[i])

			if (state == "unknown") then
				description = feature .. " is in an " .. state .. " state"
			else
				description = feature .. description .. state
			end

			RecordInfoEvent("System", "Feature", description)
		end
	end
end

--[[=============================================================================
    Sent in response to a Dynamic Data Refresh Request.

    Format: [PN][AN][TM][ESL][ESH]
===============================================================================]]
function DEV_MSG.Temperature(command)
	local buffer = { command }

	LogTrace("DEV_MSG.Temperature(): reply[%s]", command)

	local partitionID = PullIntValue(buffer)
	local areaNumber = PullIntValue(buffer)

	local currentTemp = PullIntValue(buffer)
	local lowSetpoint = PullIntValue(buffer)
	local highSetpoint = PullIntValue(buffer)

	local description = string.format("Current Temperature[%d]     Low Setpoint[%d]     High Setpoint[%d]", currentTemp, lowSetpoint, highSetpoint)

	RecordInfoEvent("System", "Temperature", description)
end

--[[=============================================================================
    Sent in response to a Dynamic Data Refresh Request(20h).

    Format: [HR][MN][MM][DD][YY]
===============================================================================]]
function DEV_MSG.TimeDate(command)
	LogTrace("DEV_MSG.TimeDate(): reply[%s]", command)
	-- NOTE: We don't currently do anything with the time/date
end

--[[=============================================================================
    Sent whenever a light state change occurs or in response to a Refresh command.

    Format: [PN][AN][LS1][LS2]
===============================================================================]]
function DEV_MSG.LightState(command)
	-- LogTrace("DEV_MSG.LightState(): reply[%s]", command)
	-- NOTE: Not used if we used it we would be spamming the history agent with unimportant data.
end

--[[=============================================================================
    Sent whenever a user lights control is made.

    Format: [PN][AN][ST][SNh][SNm][SNl][LN][LS]

    Source Type
    ------------------------------------------------------------
    0 = bus device
    1 = phone
    2 = zone
    3 = system

    Light Number
    ------------------------------------------------------------
    0 = all lights
    1-32 specific light

    Light State
    ------------------------------------------------------------
    0 = off
    1 = on
    3 = bright
    5 = dim
===============================================================================]]
function DEV_MSG.UserLights(command)
	local buffer = { command }

	LogTrace("DEV_MSG.UserLights(): reply[%s]", command)

	local partitionID = PullIntValue(buffer)
	local areaNumber = PullIntValue(buffer)

	-- The next three values are not needed for what we are doing
	PullValue(buffer)
	PullValue(buffer)
	PullValue(buffer)

	local lightNumber = PullIntValue(buffer)
	local lightState = PullIntValue(buffer)

	local description
	if (lightNumber == 0) then
		description = "All lights have been turned " .. lightStateString[lightState]
	else
		description = "Light number " .. lightNumber .. " has been turned " .. lightStateString[lightState]
	end

	RecordInfoEvent("Light", lightStateString[lightState], description)
end

--[[=============================================================================
    Sent whenever the panel receives a keypress from a keyfob.

    Format: [PN][AN][ZNh][ZNl][KC]

    KeyCode
    ------------------------------------------------------------
    0 = disarm
    1 = arm
    2 = lights
    3 = star
    4 = arm&disarm
    5 = lights&star
    6 = long lights
    9 = arm&star
    10 = disarm&lights
===============================================================================]]
function DEV_MSG.Keyfob(command)
	local buffer = { command }

	LogTrace("DEV_MSG.Keyfob(): reply[%s]", command)

	local partitionID = PullIntValue(buffer)
	local areaNumber = PullIntValue(buffer)

	-- Unused value remove from buffer
	PullValue(buffer)

	local zoneID = partitionID*100+PullIntValue(buffer)
	local keyCode = PullIntValue(buffer)

	RecordInfoEvent("System", "Keyfob", keyfobCommand[keyCode])
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--  The following functions are internal to the driver
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[=============================================================================
    Returns the checksum for the given data value
===============================================================================]]
function GetCheckSum(data)
	local i = 1
	local current = 0
	local computedSum = 0
	local length = string.len(data)

	while (i <= length) do

		current = string.sub(data, i, i + 1)
		computedSum =  computedSum + tonumber(string.format("0x%s", current))

		i = i + 2
	end

	computedSum = string.format("%02X", computedSum)
	computedSum = string.sub(computedSum, string.len(computedSum) - 1)
	-- LogDebug("GetCheckSum(): computedSum[%s]", tostring(computedSum))

	return computedSum
end

--[[=============================================================================
    Return a table of bit values for the passed in byte
===============================================================================]]
function GetBits(value, length)
	local bitCount = length or 1
	local bitString = { }
	local i = 1

	value = tonumber(value, 16)
	while (value > 0) do
		local last = value % 2
		if (last == 1) then
			bitString[i] = 1
		else
			bitString[i] = 0
		end

		value = (value - last) / 2
		i = i + 1
	end

	while (i <= bitCount) do
		bitString[i] = 0
		i = i + 1
	end

	return bitString
end

--[[=============================================================================
    Returns true or false based on the value being 1 or 0
===============================================================================]]
function IsValueTrue(value)
	value = tostring(value)
	if (value == "1" or string.lower(value) == "true") then
		return true
	end

	return false
end

--[[=============================================================================
    Updates the zone with the passed in state
===============================================================================]]
function UpdateZoneState(zoneID, zoneState)
	local status = GetBits(zoneState, 5)
	local description

	if (status[1] ~= nil) then
		if (IsZoneOpen(zoneID) ~= IsValueTrue(status[1])) then
			SetZoneState(zoneID, IsValueTrue(status[1]), false)
			if (IsValueTrue(status[1]) == true) then
				description = "Zone" .. zoneID .. " has been faulted"
			else
				description = "Zone" .. zoneID .. " has been unfaulted"
			end

			RecordInfoEvent("Zone", "Fault", description)
		end
	end

	if (status[5] ~= nil) then
		if (IsZoneBypassed(zoneID) ~= IsValueTrue(status[5])) then
			SetZoneBypassState(zoneID, IsValueTrue(status[5]), false)
			if (IsValueTrue(status[5]) == true) then
				description = "Zone" .. zoneID .. " has been bypassed"
			else
				description = "Zone" .. zoneID .. " has been unbypassed"
			end

			RecordInfoEvent("Zone", "Fault", description)
		end
	end
end

--[[=============================================================================
    Updates the partition status to the specified arm level
===============================================================================]]
function UpdatePartitionStatus(partitionID, armLevel)
	local StateType
	local PartitionState
	local CodeRequiredToArm = true
	local Description = "Partition " .. partitionID

	LogDebug("UpdatePartitionStatus(): PartitionID[%s]    ArmLevel[%s]", tostring(partitionID), tostring(armLevel))

	armLevel = tonumber(armLevel)
	if (armLevel == 1) then
		StateType = "Disarmed"
		RealArmedType = ""
		RealArmedState = ""

		PartitionState = AS_DISARMED_READY
		Description = Description .. " has been disarmed"
	elseif (armLevel > 1 and armLevel < 6) then
		PartitionState = AS_ARMED
		if (armLevel == 2) then
			StateType = "Stay"
		elseif (armLevel == 3) then
			StateType = "Away"
		elseif (armLevel == 4) then
			StateType = "Night"
		elseif (armLevel == 5) then
			StateType = "Silent"
		end

		if (IsPartitionInDelay(partitionID)) then
			RealArmedType = StateType
			RealArmedState = PartitionState
			return
		end

		Description = Description .. " has been armed as " .. StateType
	elseif (armLevel == 0) then
		StateType = "Testing"
		PartitionState = AS_DISARMED_NOT_READY
		Description = Description .. " has been set to " .. StateType
	else
		LogDebug("UpdatePartitionStatus() - Unknown armLevel[%d] specified", armLevel)
		return
	end

	RecordInfoEvent("Partition", "Arming", Description)
	if (PanelFeatures[partitionID]["Quick Arm"] == 0) then
		CodeRequiredToArm = false
	end

	SetCodeRequiredToArm(tonumber(partitionID), CodeRequiredToArm)
	SetPartitionState(tonumber(partitionID), PartitionState, StateType)
end

--[[=============================================================================
    Convert the given tokens into the correct text value
===============================================================================]]
function GetText(tokens)
	local temp
	local text = ""
	local token
	local buffer = { tokens }

	while (string.len(buffer[1]) > 0) do
		token = PullValue(buffer)
		temp = TokenValues[token]
		
		if (temp ~= nil) then
			if (string.upper(token) == "FD") then
				text = string.sub(text, 1, string.len(text)-1)
			else
				text = text .. temp
			end
		end
	end

	return text
end

--[[=============================================================================
    Print out the stored panel information
===============================================================================]]
function GetPanelInfo()
	LogDebug("DEV_MSG.PanelType(): Brand[%s],    Type[%s],    HWRevision[%s],    SWRevision[%s],    SerialNumber[%s]", GetBrand(), GetPanelType(), GetHWRevision(), GetSWRevision(), GetSerialNumber())
end

--[[=============================================================================
    Remove the next value from the given table buffer
===============================================================================]]
function PullValue(buffer)
	local param = string.sub(buffer[1], 1, 2)
	buffer[1] = string.sub(buffer[1], 3)

	return param
end

--[[=============================================================================
    Remove the next value in number format from the given table buffer
===============================================================================]]
function PullIntValue(buffer)
	local param = tonumber(string.sub(buffer[1], 1, 2), 16)
	buffer[1] = string.sub(buffer[1], 3)

	return param
end

--[[=============================================================================
    Takes the time or date value and prepends a zero on smaller values
===============================================================================]]
function PadTimeDateValue(value)
	local paddedValue = value

	if (string.len(value) == 1) then
		paddedValue = "0" .. value
	end

	return paddedValue
end

--[[=============================================================================
    Takes the key value and pads each value with a 0
===============================================================================]]
function PadKeyPresses(keys)
	local paddedKeys = ""

	for i = 1, string.len(keys) do
		paddedKeys = paddedKeys .. "0" .. string.sub(keys, i, i)
	end

	return paddedKeys
end

--[[=============================================================================
    Takes the given index and pads it with a 0 if the value is smaller than a single digit.
===============================================================================]]
function PadValue(index)
	local paddedIndex = ""

	if (string.len(index) <= 1) then
		paddedIndex = "0" .. index
	else
		paddedIndex = index
	end

	return paddedIndex
end

--[[=============================================================================
    PadZeros(data, length)

    Description
    Pads the given data with zeros up to length characters

    Parameters
    data(string)   - The string that is to be padded
    length(number) - The length of the value to be returned

    Returns
    The zero padded value
===============================================================================]]
function PadZeros(data, length)
	local response = ""
	local dataLength = string.len(data)

	length = tonumber(length)
	length = length - dataLength
	if (length <= 0) then 
		response = tostring(data)
	else
		for i = 1, length do
			response = response .. "0"
		end

		response = response .. data
	end

	return response
end

--[[=============================================================================
    Converts the given number or number in string format and converts it to its hex form.
===============================================================================]]
function ConvertToHex(num)
	local mod
	local value = ""
	local hexstr = "0123456789ABCDEF"

	num = tonumber(num)
	if (num ~= nil) then
		while num > 0 do
			mod = math.fmod(num, 16)
			value = string.sub(hexstr, mod+1, mod+1) .. value
			num = math.floor(num / 16)
		end
	end

	if (value == "") then
		value = "0"
	end

	return value
end

--[[=============================================================================
    Takes the given command and partition number and creates the Command Key Press sequence
    to pass to the panel.
===============================================================================]]
function GetCommandString(PartitionIndex, command)
	if (tonumber(PartitionIndex) == nil) then
		return ""
	end

	local lastIndex = PadValue(ConvertToHex(4 + string.len(command)))
	local request = lastIndex .. "40" .. PadKeyPresses(PartitionIndex .. "0" )
	for i = 1, string.len(command) do
		request = request .. KeypadValues[string.sub(command, i, i)]
	end

	local checkSum = GetCheckSum(request)
	LogDebug("GetCommandString:   CommandString[%s]   CheckSum[%s]", request, checkSum)

	return ("\n" .. request .. checkSum)
end

--[[=============================================================================
    Returns the text version of the features state
===============================================================================]]
function GetFeatureState(state)
	state = tonumber(state)

	if (state == 1) then
		return "enabled"
	elseif (state == nil or state ~= 0) then
		return "unknown"
	end

	return "disabled"
end

--[[=============================================================================
    Locates the given label in the TokenValues array or nil
===============================================================================]]
function FindWord(label)
	return TokenValues[label]
end

--[[=============================================================================
    Locates the given label in the ZoneLabelValues array or nil
===============================================================================]]
function FindLabel(zoneLabel)
	return ZoneLabelValues[zoneLabel]
end

--[[=============================================================================
    Will map the given interlogix zone type to a control4 zone type

    Control4 Zone Types
    ------------------------------------------------------------
    Unknown         = 0
    Contact Sensor  = 1
    Exterior Door   = 2
    Exterior Window = 3
    Interior Door   = 4
    Motion Sensor   = 5
    Fire            = 6
    Gas             = 7
    Carbon Monoxide = 8
    Heat            = 9
    Water           = 10
    Smoke           = 11
    Pressure        = 12
    Glass Break     = 13
    Gate            = 14
    Garage Door     = 15
===============================================================================]]
function GetC4ZoneType(zoneType)
	local c4ZoneType = 0

	if (zoneType == 8 or zoneType == 9 or zoneType == 21 or zoneType == 22 or zoneType == 23 or zoneType == 24) then
	   c4zoneType = 1
	elseif (zoneType == 10) then
		c4ZoneType = 2
	elseif (zoneType == 13) then
		c4ZoneType = 3
	elseif (zoneType == 14 or zoneType == 16 or zoneType == 19) then
		c4ZoneType = 4
	elseif (zoneType == 15 or zoneType == 17 or zoneType == 18 or zoneType == 20) then
		c4ZoneType = 5
	elseif (zoneType == 34) then
		c4ZoneType = 7
	elseif (zoneType == 38) then
		c4ZoneType = 10
	elseif (zoneType == 26) then
		c4ZoneType = 11
	elseif (zoneType == 28 or zoneType == 32) then
		c4ZoneType = 12
	elseif (zoneType == 12) then
		c4ZoneType = 14
	elseif (zoneType == 11) then
		c4ZoneType = 15
	end

	return c4ZoneType
end

--[[=============================================================================
    Checks to see if Start is at the beginning of String

    String(string) - The string to be checked
    Start(string)  - The value to check the string against
===============================================================================]]
function string.starts(String, Start)
	return (string.sub(String, 1, string.len(Start)) == Start)
end

--[[=============================================================================
    Checks to see if End is at the end of String

    String(string) - The string to be checked
    End(string)    - The value to check the string against
===============================================================================]]
function string.ends(String, End)
   return (End == '' or string.sub(String, -string.len(End)) == End)
end

--[[=============================================================================
    Updated the PARTITION<#>_LAST_USER and PARTITION<#>_LAST_EVENT variables
    with current event information.

    PartitionID: The partition that the event happened on
    User:        The value that the PARTITION<#>_LAST_USER variable is updated to
    Event:       The value that the PARTITION<#>_LAST_EVENT variable is updated to
===============================================================================]]
function UpdateC4Variable(PartitionID, User, Event)

	if (IsPartitionEnabled(PartitionID)) then
		C4:SetVariable("PARTITION" .. tostring(PartitionID) .. "_LAST_USER", User)
		C4:SetVariable("PARTITION" .. tostring(PartitionID) .. "_LAST_EVENT", Event)
	end
end

--[[=============================================================================
    Create the right user string based on high and low values

    high: 00 for user number 01 for keyfob
    low:  The actual user or zone number in question
===============================================================================]]
function GetUserNumber(high, low)
	if (tonumber(high) == 0) then
		return "user " .. tostring(tonumber(low))
	else
		return "zone " .. tostring(tonumber(low))
	end
end

--[[=============================================================================
    Writes an event to the history agent

    Severity:    Should be one of the following(Critical, Warning, Information)
    Category:    Should be one of the following(Security, Watch, Listen, ...)
    SubCategory: For the purposes of this driver it will be(Panel, Partition, Zone, or Light)
    EventType:   Can be anything
    Description: Should be detailed information describing what the event was
===============================================================================]]
function RecordHistoricalEvent(severity, category, subcategory, eventType, description)
	local history_agent_id = 0
	local proxydevs = C4:GetDevicesByC4iName("History.c4i")

	if (proxydevs ~= nil) then
		for k, v in pairs(proxydevs) do
			history_agent_id = k
			break
		end

		if history_agent_id > 0 then
			local event_params = {}

			event_params.SEVERITY = severity
			event_params.DEVICE_ID = C4:GetDeviceID()
			event_params.CATEGORY = category
			event_params.SUBCATEGORY = subcategory
			event_params.TYPE = eventType
			event_params.DESCRIPTION = description

			C4:SendToDevice(history_agent_id, "RECORD_HISTORICAL_EVENT", event_params)
		end
	end
end

--[[=============================================================================
    Calls the RecordHistoricalEvent function with the severity set to "Critical" and the
    category set to "Security".
===============================================================================]]
function RecordCriticalEvent(subcategory, eventType, description)
	RecordHistoricalEvent("Critical", "Security", subcategory, eventType, description)
end

--[[=============================================================================
    Calls the RecordHistoricalEvent function with the severity set to "Warning" and the
    category set to "Security".
===============================================================================]]
function RecordWarningEvent(subcategory, eventType, description)
	RecordHistoricalEvent("Warning", "Security", subcategory, eventType, description)
end

--[[=============================================================================
    Calls the RecordHistoricalEvent function with the severity set to "Information" and the
    category set to "Security".
===============================================================================]]
function RecordInfoEvent(subcategory, eventType, description)
	RecordHistoricalEvent("Info", "Security", subcategory, eventType, description)
end