--[[=============================================================================
    Functions dealing with the management of panel information, zones and their
    states, as well as pgms and their states

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
TEMPLATE_VERSION.securitypanel = "6"

--[[=============================================================================
    IsPgmValid(PgmID)

    Description: 
    Identifies whether or not the given PgmID has been added to the system

    Parameters:
    PgmID(int) - The number for the pgm in question

    Returns:
    A boolean indicating the validity of the specified pgm
===============================================================================]]
function IsPgmValid(PgmID)
	return (PgmInfoList[tonumber(PgmID)] ~= nil)
end

--[[=============================================================================
    SetPgmState(PgmID, IsOpen, Initializing)

    Description: 
    Sets the state of the specified pgm

    Parameters:
    PgmID(int)         - The number for the pgm whose state is being set
    IsOpen(bool)       - Indicates the state of the specified pgm
    Initializing(bool) - Indicates whether this is the initialization of the
                         pgm. If true then the programming events within the 
                         system will not be fired.

    Returns:
    None
===============================================================================]]
function SetPgmState(PgmID, IsOpen, Initializing)
	if (not IsPgmValid(PgmID)) then
		AddPgm(PgmID)
	end

	PgmInfoList[PgmID]:SetPgmState(IsOpen, Initializing)
end

--[[=============================================================================
    IsPgmOpen(PgmID)

    Description: 
    Identifies whether or not the specified zone is open

    Parameters:
    PgmID(int) - The number for the pgm whose open state is in question

    Returns:
    A boolean indicating the open state of the given pgm
===============================================================================]]
function IsPgmOpen(PgmID)
	if (not IsPgmValid(PgmID)) then
		return false
	end
	
	return PgmInfoList[PgmID]:GetPgmState()
end

--[[=============================================================================
    AddPgm(PgmID)

    Description: 
    Adds the specified Pgm to the managed list

    Parameters:
    PgmID(int) - The number of the pgm that is being added
===============================================================================]]
function AddPgm(PgmID)
	TheSecurityPanel:AddPgm(PgmID)
end

--[[=============================================================================
    RemovePgm(PgmID)

    Description: 
    Removes the specified Pgm from the managed list

    Parameters:
    PgmID(int) - The number of the pgm that is being removed
===============================================================================]]
function RemovePgm(PgmID)
	TheSecurityPanel:RemovePgm(PgmID)
end

--[[=============================================================================
    SetDefaultZoneName(ZoneName)

    Description: 
    Set the default label for the zone in the system, when no label is given

    Parameters:
    ZoneName(string) - The label to use for unnamed zones

    Returns:
    None
===============================================================================]]
function SetDefaultZoneName(ZoneName)
	DefaultZoneName = tostring(ZoneName)
end

--[[=============================================================================
    IsZoneValid(ZoneID)

    Description: 
    Identifies whether or not the given ZoneID has been added to the system

    Parameters:
    ZoneID(int) - The number for the zone in question

    Returns:
    A boolean indicating the validity of the specified zone
===============================================================================]]
function IsZoneValid(ZoneID)
	return (ZoneInfoList[tonumber(ZoneID)] ~= nil)
end

--[[=============================================================================
    GetTotalPgmCount()

    Description: 
    Gets the current number of PGMs that have been created by the driver

    Returns:
    Returns the current number of PGMs that have been created by the driver
===============================================================================]]
function GetTotalPgmCount()
	return #PgmInfoList
end

--[[=============================================================================
    GetTotalZoneCount()

    Description: 
    Gets the current number of zones that have been created by the driver

    Returns:
    Returns the current number of zones that havee been created by the driver
===============================================================================]]
function GetTotalZoneCount()
	return #ZoneInfoList
end

--[[=============================================================================
    IsZoneBypassed(ZoneID)

    Description: 
    Identifies whether or not the specified zone id has been bypassed

    Parameters:
    ZoneID(int) - The number for the zone whose bypass state is in question

    Returns:
    A boolean indicating the bypass state of the given zone
===============================================================================]]
function IsZoneBypassed(ZoneID)
	if (IsZoneValid(ZoneID)) then
		return ZoneInfoList[tonumber(ZoneID)]:IsBypassed()
	else
		return false
	end
end

--[[=============================================================================
    IsZoneOpen(ZoneID)

    Description: 
    Identifies whether or not the specified zone is open

    Parameters:
    ZoneID(int) - The number for the zone whose open state is in question

    Returns:
    A boolean indicating the open state of the given zone
===============================================================================]]
function IsZoneOpen(ZoneID)
	if (IsZoneValid(ZoneID)) then
		return ZoneInfoList[tonumber(ZoneID)]:IsOpen()
	else
		return false
	end
end

--[[=============================================================================
    GetZoneType(ZoneID)

    Description: 
    Identifies the type of the zone specified

    Parameters:
    ZoneID(int) - The number for the zone whose type is in question

    Returns:
    The panels zone type
===============================================================================]]
function GetZoneType(ZoneID)
	if (IsZoneValid(ZoneID)) then
		return ZoneInfoList[tonumber(ZoneID)]:GetZoneType()
	else
		return 0
	end
end

--[[=============================================================================
    SetZoneInfo(ZoneID, ZoneName, ZoneTypeID, ZoneTypeID_C4)

    Description: 
    Set the details for the specified zones

    Parameters:
    ZoneID(int)       - The number for the zone whose state is being set
    ZoneName(string)  - The identifying label of the zone
    ZoneTypeID(int)   - The zone type from the perspective of the vendor
    C4ZoneTypeID(int) - The control4 zone type that maps the vendor specific 
                        type with the control4 sensor types.

    Control4 Zone Types
    ==========================
    Unknown         = 0
    Contact Sensor  = 1
    Exterior Door   = 2
    Exterior Window = 3
    Interior Door   = 4
    Motion Sensor   = 5
    Fire            = 6
    Gas             = 7
    Carbon Monoxide = 8
    Heat            = 9
    Water           = 10
    Smoke           = 11
    Pressure        = 12
    Glass Break     = 13
    Gate            = 14
    Garage Door     = 15
===============================================================================]]
function SetZoneInfo(ZoneID, ZoneName, ZoneTypeID, ZoneTypeID_C4)
	local nZoneID = tonumber(ZoneID)

	if (not IsZoneValid(nZoneID)) then
		AddZone(nZoneID)
	end

	return ZoneInfoList[nZoneID]:SetZoneInfo(ZoneName, ZoneTypeID, ZoneTypeID_C4)
end

--[[=============================================================================
    AddZone(ZoneID)

    Description:
    Adds the specified zone number to the list of managed/monitored zones

    Parameters:
    ZoneID(int) - The number for the zone that is being added
===============================================================================]]
function AddZone(ZoneID)
	TheSecurityPanel:AddZone(ZoneID)
end

--[[=============================================================================
    RemoveZone(ZoneID)

    Description:
    Removes the specified zone number from the list of managed/monitored zones

    Parameters:
    ZoneID(int) - The number for the zone that is being removed
===============================================================================]]
function RemoveZone(ZoneID)
	TheSecurityPanel:RemoveZone(ZoneID)
end

--[[=============================================================================
    SetZoneState(ZoneID, IsOpen, Initializing)

    Description: 
    Sets the specified zones state with the system

    Parameters:
    ZoneID(int)        - The number for the zone whose state is being set
    IsOpen(bool)       - Indicates the state of the specified zone
    Initializing(bool) - Indicates whether this is the initialization of the
                         zone. If true then the programming events within the 
                         system will not be fired.

    Returns:
    True if something changed, otherwise False
===============================================================================]]
function SetZoneState(ZoneID, IsOpen, Initializing)
	if (ZoneInfoList[tonumber(ZoneID)] ~= nil) then
		return ZoneInfoList[tonumber(ZoneID)]:SetZoneState(IsOpen, Initializing)
	else
		return false
	end
end

--[[=============================================================================
    SetZoneBypassState(ZoneID, IsBypassed, Initializing)

    Description: 
    Sets the specified zones bypass state with the system

    Parameters:
    ZoneID(int)        - The number for the zone whose state is being set
    IsBypassed(bool)   - Indicates whether the zone has been bypassed for the
                         specified zone
    Initializing(bool) - Indicates whether this is the initialization of the
                         zone. If true then the programming events within the 
                         system will not be fired.

    Returns:
    True if something changed, otherwise False
===============================================================================]]
function SetZoneBypassState(ZoneID, IsBypassed, Initializing)
	if (ZoneInfoList[tonumber(ZoneID)] ~= nil) then
		return ZoneInfoList[tonumber(ZoneID)]:SetBypassState(IsBypassed, Initializing)
	else
		return false
	end
end

--[[=============================================================================
    StartTroubleCondition(TroubleMessage)

    Description: 
    Sets the given string as a trouble condition with the panel

    Parameters:
    TroubleMessage(string) - The trouble condition to set for the panel

    Returns:
    An identifier to uniquely identify this trouble condition with the panel.
===============================================================================]]
function StartTroubleCondition(TroubleMessage)
	return TheSecurityPanel:TroubleStart(TroubleMessage)
end

--[[=============================================================================
    ClearTroubleCondition(Identifier)

    Description: 
    Clears the trouble condition with the panel

    Parameters:
    Identifier(string) - An identifier to uniquely identify this trouble
                         condition with the panel.
===============================================================================]]
function ClearTroubleCondition(Identifier)
	TheSecurityPanel:TroubleClear(Identifier)
end

--[[=============================================================================
    function PanelRequestAdditionalInfo(Prompt, CurrentInfoStr, FunctionName, MaskData, InterfaceID)

    Description: 
    Provides a mechanism to ask the UI for the security panel to provide more
    info, the most common use case would be if a user code or an installers code
    would be required to complete a desired action.

    Parameters:
    Prompt(string)            - The prompt that the UI will display when asking
                                for the additional information
    CurrentInfoString(string) - A string of current information that will be
                                passed along as the new information is requested
                                as to what the driver should do with the new
                                information.  Usually it would indicate which
                                routine should be called and what parameters
                                should be passed to that routine.
    FunctionName(string)      - Optional name of a function that the handling
                                routine should call when the additional info
                                is provided
    MaskData(boolean)         - Indicates whether or not the UI should hide the
                                data (i.e. just print asterisks for letters) as
                                the requested data is entered
    InterfaceID(string)       - Unique id of the interface that originally made
                                the request
===============================================================================]]
function PanelRequestAdditionalInfo(Prompt, CurrentInfoStr, FunctionName, MaskData, InterfaceID)
	TheSecurityPanel:RequestAdditionalInfo(Prompt, CurrentInfoStr, FunctionName, MaskData, InterfaceID)
end
