--[[=============================================================================
    ZoneInformation Class

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "lib.c4_object"

TEMPLATE_VERSION.securitypanel = "6"

-- Control4 zone type mapping
ZoneTypes = {
	-- Name to ID mappings
	UNKNOWN = 0,
	CONTACT_SENSOR = 1,
	EXTERIOR_DOOR = 2,
	EXTERIOR_WINDOW = 3,
	INTERIOR_DOOR = 4,
	MOTION_SENSOR = 5,
	FIRE = 6,
	GAS = 7,
	CARBON_MONOXIDE = 8,
	HEAT = 9,
	WATER = 10,
	SMOKE = 11,
	PRESSURE = 12,
	GLASS_BREAK = 13,
	GATE = 14,
	GARAGE_DOOR = 15,
	TYPES_COUNT = 16,        -- The number of defined values
	-- ID to Name mappings
	[0] = "UNKNOWN",
	"CONTACT_SENSOR",
	"EXTERIOR_DOOR",
	"EXTERIOR_WINDOW",
	"INTERIOR_DOOR",
	"MOTION_SENSOR",
	"FIRE",
	"GAS",
	"CARBON_MONOXIDE",
	"HEAT",
	"WATER",
	"SMOKE",
	"PRESSURE",
	"GLASS_BREAK",
	"GATE",
	"GARAGE_DOOR" }

ZoneInfoList = {}
ZoneInformation = inheritsFrom(nil)
DefaultZoneName = "Sensor "

--[[=============================================================================
    Functions that are meant to be private to the class
===============================================================================]]
function ZoneInformation:construct(ZoneID)
	LogDebug("ZoneInformation:construct")
	self._ZoneID = ZoneID
	self._IsOpen = false
	self._IsBypassed = false
	self._ZoneName = DefaultZoneName .. tostring(ZoneID)
	self._ZoneTypeID_C4 = 0     -- Control4 zone type
	self._ZoneTypeID_3P = 0     -- Security panel zone type
	self._NeedToSendInitialInfo = true

	self._PartitionMemberList = {}

	ZoneInfoList[ZoneID] = self
	--self:ZoneInfoChanged()		-- force notification
end

function ZoneInformation:destruct()
	LogDebug("ZoneInformation:destruct")
	NOTIFY.PANEL_REMOVE_ZONE(self._ZoneID, TheSecurityPanel._BindingID)
	ZoneInfoList[self._ZoneID] = nil
end

function ZoneInformation:ZonePanelXML()
	LogDebug("ZoneInformation:ZonePanelXML")
	local ZoneXMLInfo = {}

	table.insert(ZoneXMLInfo, MakeXMLNode("id", tostring(self._ZoneID)))
	table.insert(ZoneXMLInfo, MakeXMLNode("name", tostring(self._ZoneName)))
	table.insert(ZoneXMLInfo, MakeXMLNode("type_id", tostring(self._ZoneTypeID_C4)))
	table.insert(ZoneXMLInfo, MakeXMLNode("partitions", self:ListPartitions()))
	table.insert(ZoneXMLInfo, MakeXMLNode("can_bypass", "true"))
	table.insert(ZoneXMLInfo, MakeXMLNode("is_open", tostring(self._IsOpen)))

	return MakeXMLNode("zone", table.concat(ZoneXMLInfo, "\n"))
end

function ZoneInformation:ZonePartitionXML()
	LogDebug("ZoneInformation:ZonePartitionXML")
	return MakeXMLNode("zone", MakeXMLNode("id", tostring(self._ZoneID)))
end

function ZoneInformation:ListPartitions()
	LogDebug("ZoneInformation:ListPartitions")
	local PartitionList = ""

	for i,p in pairs(self._PartitionMemberList) do
		if (PartitionList ~= "") then
			PartitionList = (PartitionList .. ",")
		end
		
		PartitionList = PartitionList .. tostring(i)
	end

	return (PartitionList ~= "") and PartitionList or " "
end

function ZoneInformation:ZoneInfoChanged()
	LogDebug("ZoneInfoChanged => Zone %d: Name = %s, Type = %d, Type_Alt = %d", tonumber(self._ZoneID), tostring(self._ZoneName), tonumber(self._ZoneTypeID_C4), tonumber(self._ZoneTypeID_3P))
	NOTIFY.PANEL_ZONE_INFO(self._ZoneID, self._ZoneName, self._ZoneTypeID_C4, self:ListPartitions(), self._IsOpen, TheSecurityPanel._BindingID)
end

--[[=============================================================================
    Functions that are wrappered and meant to be exposed to the driver
===============================================================================]]
function ZoneInformation:IsBypassed()
	LogDebug("ZoneInformation:IsBypassed")
	return self._IsBypassed
end

function ZoneInformation:IsOpen()
	LogDebug("ZoneInformation:IsOpen")
	return self._IsOpen
end

function ZoneInformation:GetZoneType()
	LogDebug("ZoneInformation:GetZoneType")
	return self._ZoneTypeID_3P
end

function ZoneInformation:GetZoneState()
	LogDebug("ZoneInformation:GetZoneState")
	return self._IsOpen and "OPENED" or "CLOSED"
end

function ZoneInformation:SetZoneInfo(ZoneName, ZoneTypeID, ZoneTypeID_C4)
	LogDebug("ZoneInformation:SetZoneInfo")
	local SomethingChanged = false

	if (self._ZoneName ~= ZoneName and ZoneName ~= "") then
		self._ZoneName = ZoneName
		SomethingChanged = true
	end

	if (self._ZoneTypeID_C4 ~= ZoneTypeID_C4 and ZoneTypeID_C4 ~= "") then
		self._ZoneTypeID_C4 = ZoneTypeID_C4
		SomethingChanged = true
	end

	if (self._ZoneTypeID_3P ~= ZoneTypeID and ZoneTypeID ~= "") then
		self._ZoneTypeID_3P = ZoneTypeID
	end

	if (SomethingChanged) then
		self:ZoneInfoChanged()
	end
end

function ZoneInformation:AddToPartition(PartitionID)
	LogDebug("ZoneInformation:AddToPartition")
	self._PartitionMemberList[PartitionID] = SecurityPartitionIndexList[PartitionID]
end

function ZoneInformation:RemoveFromPartition(PartitionID)
	LogDebug("ZoneInformation:RemoveFromPartition")
	self._PartitionMemberList[PartitionID] = nil
end

function ZoneInformation:SetZoneState(IsOpen, Initializing)
	LogDebug("ZoneInformation:SetZoneState")
	local RetVal = false     -- return true if the state changed, false if it didn't
	local JustInitializing = Initializing or false

	if ((self._IsOpen ~= IsOpen) or self._NeedToSendInitialInfo or JustInitializing) then
		self._IsOpen = IsOpen
		RetVal = true

		LogDebug("!!!!!!   Zone %d %s  !!!!!!", tonumber(self._ZoneID), tostring(self:GetZoneState()))
		for k,CurPartition in pairs(self._PartitionMemberList) do
			NOTIFY.ZONE_STATE(self._ZoneID, self._IsOpen, self._IsBypassed, CurPartition._BindingID)
		end
 
		NOTIFY.PANEL_ZONE_STATE(self._ZoneID, self._IsOpen, self._NeedToSendInitialInfo or JustInitializing, TheSecurityPanel._BindingID)
		self._NeedToSendInitialInfo = false
	end

	return RetVal
end

function ZoneInformation:SetBypassState(IsBypassed, Initializing)
	LogDebug("ZoneInformation:SetBypassState")
	local RetVal = false     -- return true if the state changed, false if it didn't
	local JustInitializing = Initializing or false

	if (self._IsBypassed ~= IsBypassed) then

		RetVal = true
		self._IsBypassed = IsBypassed

		if (not JustInitializing) then

			for k,CurPartition in pairs(self._PartitionMemberList) do
				NOTIFY.ZONE_STATE(self._ZoneID, self._IsOpen, self._IsBypassed, CurPartition._BindingID)
			end
		end
	end

	return RetVal
end