--[[=============================================================================
    SecurityPartition Class

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_utils"
require "lib.c4_object"
require "lib.c4_log"
require "partition_proxy.securitypartition_functions"
require "partition_proxy.securitypartition_commands"
require "partition_proxy.securitypartition_notifies"
require "panel_proxy.zone_info"

TEMPLATE_VERSION.securitypartition = "9"

SecurityPartitionIndexList = {}
SecurityPartitionFunctions = {}
SecurityPartitionBindingList = {}

--[[=============================================================================
    ARMED STATES
===============================================================================]]
AS_ARMED = "ARMED"
AS_ALARM = "ALARM"
AS_EXIT_DELAY = "EXIT_DELAY"
AS_ENTRY_DELAY = "ENTRY_DELAY"
AS_DISARMED_READY = "DISARMED_READY"
AS_DISARMED_NOT_READY = "DISARMED_NOT_READY"

SecurityPartition = inheritsFrom(nil)

--[[=============================================================================
    Functions that are meant to be private to the class
===============================================================================]]
function SecurityPartition:construct(PartitionNumber, BindingID)
	self._IsEnabled = true
	self._PartitionNumber = PartitionNumber
	self._BindingID = BindingID
	self._CurrentPartitionState = "Unknown"
	self._InitializingStatus = false
	self._MyZoneList = {}

	self._DelayTimeTotal = 0
	self._DelayTimeRemaining = 0
	self._OpenZoneCount = 0

	self._CodeRequiredToArm = false
	self._CodeRequiredToClear = true
	self._InAlarm = false
	self._CurrentStateType = ""
	self._DefaultUserCode = ""

	SecurityPartitionIndexList[PartitionNumber] = self
	SecurityPartitionBindingList[BindingID] = self
end

function SecurityPartition:InitialSetup()
end

function SecurityPartition:SetInitializingFlag(FlagValue)
	self._InitializingStatus = FlagValue
end

function SecurityPartition:EntryExitDelay(DelayType, DelayActive, TotalTime, RemainingTime)
	local DelayMessage = DelayType .. " Delay"
	local OldTotalTime = self._DelayTimeTotal
	local OldRemainingTime = self._DelayTimeRemaining

	if (DelayActive) then
		-- Delay On
		if (TotalTime == 0) then
			TotalTime = OldTotalTime
		end

		self._DelayTimeTotal = (TotalTime > RemainingTime) and TotalTime or RemainingTime
		self._DelayTimeRemaining = RemainingTime
		DelayMessage = DelayMessage .. " On"
	else
		-- Delay Off
		self._DelayTimeTotal = 0
		self._DelayTimeRemaining = 0
		DelayMessage = DelayMessage .. " Off"
	end

	LogDebug(DelayMessage)
	return ((OldTotalTime ~= self._DelayTimeTotal) or (OldRemainingTime ~= self._DelayTimeRemaining))
end

function SecurityPartition:PartitionXML()
	local PartitionXMLInfo = {}

	table.insert(PartitionXMLInfo, MakeXMLNode("id", tostring(self._PartitionNumber)))
	table.insert(PartitionXMLInfo, MakeXMLNode("enabled", tostring(self._IsEnabled)))
	table.insert(PartitionXMLInfo, MakeXMLNode("binding_id", tostring(self._BindingID)))
	table.insert(PartitionXMLInfo, MakeXMLAttrNode("state", tostring(self._CurrentPartitionState), "type", tostring(self._CurrentStateType)))

	return MakeXMLNode("partition", table.concat(PartitionXMLInfo, "\n"))
end

function SecurityPartition:NotifyPartitionState()

	LogTrace("Partition %d set to partition state %s : %s Alarm is %s", tonumber(self._PartitionNumber), tostring(self._CurrentPartitionState), tostring(self._CurrentStateType), tostring(self._InAlarm))
	if (self._InitializingStatus) then
		NOTIFY.PARTITION_STATE_INIT(self._CurrentPartitionState, self._CurrentStateType, self._DelayTimeTotal, self._DelayTimeRemaining, self._CodeRequiredToClear, self._BindingID)
	else
		NOTIFY.PARTITION_STATE(self._CurrentPartitionState, self._CurrentStateType, self._DelayTimeTotal, self._DelayTimeRemaining, self._CodeRequiredToClear, self._BindingID)
	end

	NOTIFY.PANEL_PARTITION_STATE(self._PartitionNumber, self._CurrentPartitionState, self._CurrentStateType, PANEL_PROXY_BINDINGID)
end

function SecurityPartition:ClearDisplayText()
	NOTIFY.DISPLAY_TEXT("", self._BindingID)
end

--[[=============================================================================
    Functions for handling request from the Partition Proxy
===============================================================================]]
function SecurityPartition:PrxGetCurrentState(tParams)
	LogTrace("GetCurrentState for partition %d", tonumber(self._PartitionNumber))
	NOTIFY.PARTITION_STATE_INIT(self._CurrentPartitionState, self._CurrentStateType, self._DelayTimeTotal, self._DelayTimeRemaining, self._CodeRequiredToClear, self._BindingID)
end

function SecurityPartition:PrxPartitionArm(tParams)
	local ArmType = tParams["ArmType"]
	local UserCode = tParams["UserCode"]
	local Bypass = toboolean(tParams["Bypass"])
	local InterfaceID = tParams["InterfaceID"] or ""

	if (InterfaceID == "DirectorProgramming") then
		UserCode = self:GetDefaultUserCode()
	end
	
     if (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode == UserCode) and (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode ~= "") then
	   LogTrace("PrxPartitionArm %d %s %s %s", tonumber(self._PartitionNumber), tostring(ArmType), tostring(UserCode), tostring(Bypass))
	   SecCom_SendArmPartition(self._PartitionNumber, ArmType, UserCode, Bypass, InterfaceID)
     elseif (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode ~= UserCode) and (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode ~= "") then
	   LogTrace("Please enter the correct user code !!!")
     elseif (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode ~= UserCode) and (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode == "") then
	   LogTrace("Please set user code firstly !!!")
     end
end

function SecurityPartition:PrxPartitionDisarm(tParams)
	local UserCode = tParams["UserCode"]
	local InterfaceID = tParams["InterfaceID"] or ""

	if (InterfaceID == "DirectorProgramming") then
		UserCode = self:GetDefaultUserCode()
	end
	   
     if (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode == UserCode) and (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode ~= "") then
	   LogTrace("PartitionDisarm")
	   SecCom_SendDisarmPartition(self._PartitionNumber, UserCode, InterfaceID)
     elseif (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode ~= UserCode) and (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode ~= "") then
	   LogTrace("Please enter the correct user code !!!")
     elseif (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode ~= UserCode) and (SecurityPartitionIndexList[self._PartitionNumber]._DefaultUserCode == "") then
	   LogTrace("Please set user code firstly !!!")
     end
end

function SecurityPartition:PrxArmCancel(tParams)
	local InterfaceID = tParams["InterfaceID"] or ""

	LogTrace("ArmCancel")
	SecCom_ArmCancel(self._PartitionNumber, InterfaceID)
end

function SecurityPartition:PrxExecuteEmergency(tParams)
	local EmergencyType = tParams["EmergencyType"]
	local InterfaceID = tParams["InterfaceID"] or ""

	LogTrace("ExecuteEmergency")
	SecCom_SendExecuteEmergency(self._PartitionNumber, EmergencyType, InterfaceID)
end

function SecurityPartition:PrxExecuteFunction(tParams)
	local FunctionName = tParams["Function"]
	local InterfaceID = tParams["InterfaceID"]
	local trimmedCommand = string.gsub(FunctionName, " ", "")

	LogTrace("Execute Function (%s)", tostring(FunctionName))
	if (SecurityPartitionFunctions[FunctionName] ~= nil and type(SecurityPartitionFunctions[FunctionName]) == "function") then
		SecurityPartitionFunctions[FunctionName](self._PartitionNumber, InterfaceID)
	elseif (SecurityPartitionFunctions[trimmedCommand] ~= nil and type(SecurityPartitionFunctions[trimmedCommand]) == "function") then
		SecurityPartitionFunctions[trimmedCommand](self._PartitionNumber, InterfaceID)
	else
		LogInfo("ID specified is null or not a function name[%s]", tostring(FunctionName))
	end
end

function SecurityPartition:PrxKeyPress(tParams)
	local KeyName = tParams["KeyName"]
	local InterfaceID = tParams["InterfaceID"] or ""

	LogTrace("PrxKeyPress")
	SecCom_SendKeyPress(self._PartitionNumber, KeyName, InterfaceID)
end

function SecurityPartition:PrxAdditionalInfo(tParams)
	local fxName = { }
	local NewInfo = tParams["NewInfo"]
	local ParmString = tParams["InfoString"]
	local InterfaceID = tParams["InterfaceID"] or ""
	local FunctionName = tParams["FunctionName"]

	if (string.len(tostring(ParmString)) > 0) then
		ParmString = ParmString .. ","
	end
	ParmString = ParmString .. tostring(NewInfo)

	local pos = string.find(FunctionName, "%.")
	if (pos > 0) then
		fxName[1] = string.sub(FunctionName, 1, pos-1)
		fxName[2] = string.sub(FunctionName, pos+1)

		_G[fxName[1]][fxName[2]](unpack(StringSplit(ParmString, ",")))
	else
		_G[FunctionName](unpack(StringSplit(ParmString, ",")))
	end
end

function SecurityPartition:PrxSetDefaultUserCode(tParams)
	local NewUserCode = tParams["Code"]
     
	LogTrace("PrxSetDefaultUserCode  Code is >>%s<<", NewUserCode)
	self:SetDefaultUserCode(NewUserCode)
end


--[[=============================================================================
    Functions that are wrappered and meant to be exposed to the driver
===============================================================================]]
function SecurityPartition:RequestAdditionalInfo(Prompt, InfoStr, FunctionName, MaskData, InterfaceID)
	NOTIFY.REQUEST_ADDITIONAL_INFO(Prompt, InfoStr, FunctionName, MaskData, InterfaceID, self._BindingID)
end

function SecurityPartition:ArmFailed(Action, InterfaceID)
	NOTIFY.ARM_FAILED(Action, InterfaceID, self._BindingID)
end

function SecurityPartition:DisarmFailed(InterfaceID)
	NOTIFY.DISARM_FAILED(InterfaceID, self._BindingID)
end

function SecurityPartition:GetZoneIDs()
	local i = 1
	local s = {}

	for k, _ in pairs(self._MyZoneList) do
		s[i] = k
		i = i + 1
	end

	return s
end

function SecurityPartition:GetZoneCount()
	local i = 0

	for _, _ in pairs(self._MyZoneList) do
		i = i + 1
	end

	return i
end

function SecurityPartition:ContainsZone(ZoneID)
	for _, v in pairs(self._MyZoneList) do
		if(v._ZoneID == ZoneID) then
			return true
		end
	end

	return false
end

function SecurityPartition:AddZone(ZoneID)
	local nZoneID = tonumber(ZoneID)
	local TargZone = ZoneInfoList[nZoneID]

	if (TargZone == nil) then
		ZoneInformation:new(nZoneID)
		TargZone = ZoneInfoList[nZoneID]

		-- force a notification to the panel driver
		TargZone:ZoneInfoChanged()
	end

	self._MyZoneList[ZoneID] = TargZone
	NOTIFY.HAS_ZONE(ZoneID, self._BindingID)
end

function SecurityPartition:RemoveZone(ZoneID)
	local TargZone = ZoneInfoList[tonumber(ZoneID)]

	if (TargZone ~= nil) then
		self._MyZoneList[ZoneID] = nil
		NOTIFY.REMOVE_ZONE(ZoneID, self._BindingID)
	end
end

function SecurityPartition:SetEnabled(Enabled)
	if(self._IsEnabled ~= Enabled) then
		self._IsEnabled = Enabled
		NOTIFY.PARTITION_ENABLED(self._IsEnabled, self._BindingID)
	end
end

function SecurityPartition:IsEnabled()
	return self._IsEnabled
end

function SecurityPartition:SetPartitionState(NewState, NewStateType, TotalDelayTime, RemainingDelayTime, CodeRequiredToClear)
	local previousState = self._CurrentPartitionState
	local previousStateType = self._CurrentStateType
	local SendChange = ((previousState ~= NewState) or (previousStateType ~= NewStateType))
	local TotDelayTime = TotalDelayTime or 0
	local RemDelayTime = RemainingDelayTime or 0

	LogTrace("Partition %d changed state to %s : %s From %s", tonumber(self._PartitionNumber), tostring(NewState), tostring(NewStateType), tostring(self._CurrentPartitionState))
	self._CurrentPartitionState = NewState
	self._CurrentStateType = NewStateType

	if (self._CurrentPartitionState == AS_ALARM) then
		self._InAlarm = true
		self._CodeRequiredToClear = CodeRequiredToClear or false
	else
		self._InAlarm = false
		self._CodeRequiredToClear = true
	end

	if ((NewState == AS_EXIT_DELAY) or (NewState == AS_ENTRY_DELAY) or (previousState == AS_EXIT_DELAY) or (previousState == AS_ENTRY_DELAY)) then
		local DelayType = ((NewState == AS_EXIT_DELAY) or (previousState == AS_EXIT_DELAY)) and "Exit" or "Entry"
		local DelayActive = ((NewState == AS_EXIT_DELAY) or (NewState == AS_ENTRY_DELAY))
		
		if (self:EntryExitDelay(DelayType, DelayActive, TotDelayTime, RemDelayTime)) then
			SendChange = true
		end
	end

	if (SendChange) then
		self:NotifyPartitionState()
	end
end


function SecurityPartition:HaveEmergency(EmergencyType)
	NOTIFY.EMERGENCY_TRIGGERED(EmergencyType, self._BindingID)
end

function SecurityPartition:DisplayText(Message)
	NOTIFY.DISPLAY_TEXT(Message, self._BindingID)
end

function SecurityPartition:SetCodeRequiredToArm(CodeRequired)
	self._CodeRequiredToArm = CodeRequired
	NOTIFY.CODE_REQUIRED(self._CodeRequiredToArm, self._BindingID)
end

function SecurityPartition:GetPartitionState()
	return self._CurrentPartitionState
end

function SecurityPartition:GetPartitionStateType()
	return self._CurrentStateType
end

function SecurityPartition:IsArmed()
	return ((self._CurrentPartitionState == AS_ARMED) or
	        (self._CurrentPartitionState == AS_ENTRY_DELAY) or
	        (self._CurrentPartitionState == AS_ALARM))
end

function SecurityPartition:IsInDelay()
	return ((self._CurrentPartitionState == AS_EXIT_DELAY) or
	        (self._CurrentPartitionState == AS_ENTRY_DELAY))
end

function SecurityPartition:SetDefaultUserCode(NewCode)
	if (NewCode ~= nil) then
		LogTrace("Setting Default User Code for partition %d  to >>%s<<", self._PartitionNumber, NewCode)
		self._DefaultUserCode = NewCode
	end
end

function SecurityPartition:GetDefaultUserCode()
	return self._DefaultUserCode
end
