--[[=============================================================================
    Functions dealing with the management of partitions and their states

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_utils"

TEMPLATE_VERSION.securitypartition = "9"

--[[=============================================================================
    ArmPartitionFailed(PartitionID, Action, InterfaceID)

    Description: 
    Notifies the system that an arm partition has failed, and tells the UI what
    action if any needs to be taken in order to proceed.

    Parameters:
    PartitionID(int)    - The index of the partition we are arming
    Action(string)      - Indicates the action that the UI should take to help 
                          rectify. Following is a list of actions that can be
                          taken keypad(if a keycode is needed), bypass, or
                          NA(general failure)
    InterfaceID(string) - The unique identifier string of the UI which requested
                          the arming action.  We only want that UI to respond to
                          the failed message.

    Returns:
    None
===============================================================================]]
function ArmPartitionFailed(PartitionID, Action, InterfaceID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		SecurityPartitionIndexList[tonumber(PartitionID)]:ArmFailed(Action, InterfaceID)
	end
end

--[[=============================================================================
    DisarmPartitionFailed(PartitionID, InterfaceID)

    Description: 
    Notifies the system that a disarm partition has failed.

    Parameters:
    PartitionID(int)    - The index of the partition we are disarming
    InterfaceID(string) - The unique identifier string of the UI which requested
                          the disarming action. 
                          We only want that UI to respond to the failed message.

    Returns:
    None
===============================================================================]]
function DisarmPartitionFailed(PartitionID, InterfaceID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		SecurityPartitionIndexList[tonumber(PartitionID)]:DisarmFailed(InterfaceID)
	end
end

--[[=============================================================================
    GetPartitionZoneIDs(PartitionID)

    Description: 
    Get the list of zone IDs that are associated with the specified partition

    Parameters:
    PartitionID(int) - The index of the partition we are getting the list from

    Returns:
    A table containing a list of the zone numbers for the specified partition
===============================================================================]]
function GetPartitionZoneIDs(PartitionID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		return SecurityPartitionIndexList[tonumber(PartitionID)]:GetZoneIDs()
	else
		return nil
	end
end

--[[=============================================================================
    GetPartitionZoneCount(PartitionID)

    Description: 
    Get the count of the zones that are associated with the zone

    Parameters:
    PartitionID(int) - The index of the partition we are getting the count from

    Returns:
    The zone count for the associated partition
===============================================================================]]
function GetPartitionZoneCount(PartitionID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		return SecurityPartitionIndexList[tonumber(PartitionID)]:GetZoneCount()
	else
		return 0
	end
end

--[[=============================================================================
    AddZoneToPartition(PartitionID, ZoneID)

    Description: 
    Adds the given zone to the specified partition
    Note: SetZoneInfo must be called before this function in order for the call
          to succeed

    Parameters:
    PartitionID(int) - The index of the partition we are adding the zone to
    ZoneID(int)      - The zone id that is being added to the partition

    Returns:
    None
===============================================================================]]
function AddZoneToPartition(PartitionID, ZoneID)
    LogTrace("AddZoneToPartition:" .. PartitionID .. ", " .. ZoneID)
	local TargPartition = SecurityPartitionIndexList[tonumber(PartitionID)]

	if(TargPartition ~= nil) then
        gPartitionID = PartitionID
		if(not TargPartition:ContainsZone(ZoneID)) then
			TargPartition:AddZone(tonumber(ZoneID))
			ZoneInfoList[tonumber(ZoneID)]:AddToPartition(tonumber(PartitionID))
		end
	end
end

--[[=============================================================================
    RemoveZoneFromPartition(PartitionID, ZoneID)

    Description: 
    Removes the given zone from the specified partition

    Parameters:
    PartitionID(int) - The index of the partition we are adding the zone to
    ZoneID(int)      - The zone id that is being added to the partition

    Returns:
    None
===============================================================================]]
function RemoveZoneFromPartition(PartitionID, ZoneID)
	local TargPartition = SecurityPartitionIndexList[tonumber(PartitionID)]

	if(TargPartition ~= nil) then
		if(TargPartition:ContainsZone(ZoneID)) then
			TargPartition:RemoveZone(tonumber(ZoneID))
			ZoneInfoList[tonumber(ZoneID)]:RemoveFromPartition(tonumber(PartitionID))
            ZoneInfoList[tonumber(ZoneID)]:destruct()
		end
	end
end

--[[=============================================================================
    PartitionContainsZone(PartitionID, ZoneID)

    Description: 
    Tells if the given zone is a member of the given partition

    Parameters:
    PartitionID(int) - The index of the partition we are checking
    ZoneID(int)      - The id of the zone we are checking

    Returns:
    True if the partition contains the zone; False if it doesn't
===============================================================================]]
function PartitionContainsZone(PartitionID, ZoneID)
	local TargPartition = SecurityPartitionIndexList[tonumber(PartitionID)]

	if(TargPartition ~= nil) then
		return TargPartition:ContainsZone(ZoneID)
	else
		return false
	end
end

--[[=============================================================================
    SetPartitionEnabled(PartitionID, Enabled)

    Description: 
    Marks the specified partition as enabled within the system. If set to false
    the partition will not be visible to the UI.

    Parameters:
    PartitionID(int) - The index of the partition we are enabling or disabling
    Enabled(bool)    - The state of the partition

    Returns:
    None
===============================================================================]]
function SetPartitionEnabled(PartitionID, Enabled)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		SecurityPartitionIndexList[tonumber(PartitionID)]:SetEnabled(Enabled)
	end
end

--[[=============================================================================
    SetPartitionInitializingFlag(PartitionID, InitializingFlag)

    Description: 
    Marks the specified partition as currently being initialized. Some behaviours
    are different if a setting is happening for the first time

    Parameters:
    PartitionID(int)       - The index of the partition we are enabling or disabling
    InitializingFlag(bool) - True if the partition is currently being initialized

    Returns:
    None
===============================================================================]]
function SetPartitionInitializingFlag(PartitionID, InitializingFlag)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		SecurityPartitionIndexList[tonumber(PartitionID)]:SetInitializingFlag(InitializingFlag)
	end
end

--[[=============================================================================
    IsPartitionEnabled(PartitionID)

    Description: 
    Identifies whether or not the specified partition is enabled

    Parameters:
    PartitionID(int) - The index of the partition we are checking

    Returns:
    True if the partition is enabled
===============================================================================]]
function IsPartitionEnabled(PartitionID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		return SecurityPartitionIndexList[tonumber(PartitionID)]:IsEnabled()
	else
		return false
	end
end

--[[=============================================================================
    SetPartitionState(PartitionID, State, StateType, TotalDelayTime, RemainingDelayTime, CodeRequiredToClear)

    Description: 
    Sets the specified partitions current state in the system

    Parameters:
    PartitionID(int)             - The number for the partition whose state is being set
    State(string)                - The state of the partition indicated by PartitionID
                                   Following are a list of valid states(ARMED, ALARM, 
                                   DISARMED_NOT_READY, DISARMED_READY, EXIT_DELAY,
                                   and ENTRY_DELAY)
    StateType(string)            - Some description to further clarify the partition
                                   state. If the state is ARMED, the state type might
                                   be "Home" or "Away".  If the state is ALARM, the
                                   state type might be "FIRE" or "BURGLARY". This
                                   may also be an empty string for other states.
    TotalDelayTime(int)          - An optional parameter that is to be used when
                                   the state being specified is either (ENTRY_DELAY
                                   or EXIT_DELAY)
    RemainingDelayTime(int)      - An optional parameter that is to be used when
                                   the state being specified is either (ENTRY_DELAY
                                   or EXIT_DELAY)
    CodeRequiredToClear(boolean) - An optional parameter that is defaulted to true.
                                   Identifies whether or not a code is required to
                                   clear the alarm event

    Returns:
    None
===============================================================================]]
function SetPartitionState(PartitionID, State, StateType, TotalDelayTime, RemainingDelayTime, CodeRequiredToClear)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		ArmType = ArmType or ""
		TotalDelayTime = tonumber(TotalDelayTime) or 0
		RemainingDelayTime = tonumber(RemainingDelayTime) or 0

		if (CodeRequiredToClear == nil) then 
			CodeRequiredToClear = true 
		end

		SecurityPartitionIndexList[tonumber(PartitionID)]:SetPartitionState(State, StateType, TotalDelayTime, RemainingDelayTime, CodeRequiredToClear)
	end
end

--[[=============================================================================
    GetPartitionState(PartitionID)

    Description:
    Get the state of the partition that was specified by the given PartitionID

    Parameters:
    PartitionID(int) - The index of the partition we are getting the state from

    Returns:
    The state of the partition specified by the PartitionID
    Following are a list of states that should be returned (ARMED, ALARM, 
    DISARMED_NOT_READY, DISARMED_READY, EXIT_DELAY, and ENTRY_DELAY)
===============================================================================]]
function GetPartitionState(PartitionID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		return SecurityPartitionIndexList[tonumber(PartitionID)]:GetPartitionState()
	else
		return AS_DISARMED_READY
	end
end

--[[=============================================================================
    GetPartitionStateType(PartitionID)

    Description: 
    Get the state type of the partition that was specified by the given PartitionID

    Parameters:
    PartitionID(int) - The index of the partition we are getting the state type from

    Returns:
    The description of the state for the partition specified by the PartitionID
===============================================================================]]
function GetPartitionStateType(PartitionID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		return SecurityPartitionIndexList[tonumber(PartitionID)]:GetPartitionStateType()
	else
		return "UNKNOWN"
	end
end

--[[=============================================================================
    SetCodeRequiredToArm(PartitionID, CodeRequired)

    Description: 
    Tells the system that the given partition requires a code to arm.

    Parameters:
    PartitionID(int)   - The index of the partition we are specifiying the status
    CodeRequired(bool) - True if a code is required to arm the partition, and
                         false otherwise.

    Returns:
    None
===============================================================================]]
function SetCodeRequiredToArm(PartitionID, CodeRequired)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		SecurityPartitionIndexList[tonumber(PartitionID)]:SetCodeRequiredToArm(CodeRequired)
	end
end

--[[=============================================================================
    IsPartitionArmed(PartitionID)

    Description: 
    Returns the armed state of the partition indicated by PartitionID

    Parameters:
    PartitionID(int) - The index of the partition we are getting the armed
                       status for

    Returns:
    The armed state of the partition specified
===============================================================================]]
function IsPartitionArmed(PartitionID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		return SecurityPartitionIndexList[tonumber(PartitionID)]:IsArmed()
	else
		return false
	end
end

--[[=============================================================================
    IsPartitionInDelay(PartitionID)

    Description: 
    Returns the delay information for the partition indicated by PartitionID

    Parameters:
    PartitionID(int) - The index of the partition we are getting delay
                       information for

    Returns:
    True if the Partition is currently in a delay state, false otherwise
===============================================================================]]
function IsPartitionInDelay(PartitionID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		return SecurityPartitionIndexList[tonumber(PartitionID)]:IsInDelay()
	else
		return false
	end
end

--[[=============================================================================
    DisplayPartitionText(PartitionID, Message)

    Description: 
    Writes the given message to the specified partition

    Parameters:
    PartitionID(int) - The index of the partition we are writing the message to
    Message(string)  - The message to be written to the UI

    Returns:
    Writes the given message to the display field of the UI
===============================================================================]]
function DisplayPartitionText(PartitionID, Message)

	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		SecurityPartitionIndexList[tonumber(PartitionID)]:DisplayText(Message)
	end
end

--[[=============================================================================
    HaveEmergency(EmergencyName)

    Description: 
    Notifies all partitions that an emergency has been triggered.

    Parameters:
    EmergencyName(string) - The type of emergency that is being triggered.
                            Current Emergency Types: 
                            Fire, Medical, Police, and Panic.
                            However other strings could be sent if desired. The
                            UI just may not have icons for them

    Returns:
    None
===============================================================================]]
function HaveEmergency(EmergencyName)

	for _, v in pairs(SecurityPartitionIndexList) do
		v:HaveEmergency(EmergencyName)
	end
end

--[[=============================================================================
    PartitionRequestAdditionalInfo(PartitionID, Prompt, ParmList, FunctionName, MaskData, InterfaceID)

    Description: 
    Provides a mechanism to ask the UI for the given partition to provide more
    info, the most common use case would be if a user code or an installers code
    would be required to complete a desired action.

    Parameters:
    PartitionID(int)     - The index of the partition to provide the new information.
    Prompt(string)       - The prompt that the UI will display when asking for the
                           additional information.
    ParmList(string)     - A string of current information that will be passed
                           along as the new information is requested and then passed
                           back to this driver. This string would contain information
                           as to what the driver should do with the new information.
                           Usually it would indicate which routine should be called
                           and what parameters should be passed to that routine.
    FunctionName(string) - The name of the function to be called on return.
    MaskData(boolean)    - True the input off the keypad will be obscured when 
                           entering data, False the input will be visiable
    InterfaceId(string)  - A unique string identifying which interface this
                           request is being sent to. Usually this would be the
                           InterfaceId paramater given to the command that is
                           requesting more info.  An empty string is also legal
                           the command will be acted upon by all navigator 
                           interfaces.
===============================================================================]]
function PartitionRequestAdditionalInfo(PartitionID, Prompt, ParmList, FunctionName, MaskData, InterfaceId)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		SecurityPartitionIndexList[tonumber(PartitionID)]:RequestAdditionalInfo(tostring(Prompt), ParmList, tostring(FunctionName), toboolean(MaskData), tostring(InterfaceId))
	end
end

--[[=============================================================================
    SetDefaultUserCode(PartitionID, NewCode)

    Description: 
    Provides a mechanism to store a default user code for times when the user
    will be unavailable(i.e. Composer programming events)

    Parameters:
    PartitionID(int) - The index of the partition to provide the new information.
    NewCode(string)  - The default user code that will be provided when there is 
                       no user to provide the input.
===============================================================================]]
function SetDefaultUserCode(PartitionID, NewCode)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		SecurityPartitionIndexList[tonumber(PartitionID)]:SetDefaultUserCode(NewCode)
	end
end

--[[=============================================================================
    SetDefaultUserCode(PartitionID, NewCode)

    Description: 
    Returns the stored default user code

    Parameters:
    PartitionID(int) - The index of the partition to provide the new information.
===============================================================================]]
function GetDefaultUserCode(PartitionID)
	if (SecurityPartitionIndexList[tonumber(PartitionID)] ~= nil) then
		return SecurityPartitionIndexList[tonumber(PartitionID)]:GetDefaultUserCode()
	else
		return ""
	end
end
